from typing import Dict, Any

from tensorflow import keras
from nlp_tools.layers import L, KConditionalRandomField
from nlp_tools.tasks.labeling.bi_lstm_crf_model import BiLSTM_CRF_Model




class BiLstmCrfAddResidualModel(BiLSTM_CRF_Model):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_blstm': {
                'units': 64,
                'return_sequences': True
            },

            'layer_dropout': {
                'rate': 0.5
            },
            'layer_time_distributed': {},
        }


    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        crf = KConditionalRandomField()



        layer_stack = [
            L.Dropout(**config['layer_dropout'], name='layer_dropout'),
            L.Dense(output_dim, **config['layer_time_distributed']),
            crf
        ]

        dirct_output = L.Dense(config['layer_blstm']['units'] * 2)(embed_model.output)
        lstm_output  = L.Bidirectional(L.LSTM(**config['layer_blstm']), name='layer_blstm')(embed_model.output)

        # lstm output和bert output直接相加，类似与residual效果
        tensor = L.concatenate([lstm_output ,dirct_output])

        tensor = L.Dropout(**config['layer_dropout'])(tensor)
        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model = keras.Model(embed_model.inputs, tensor)
        self.crf_layer = crf













from nlp_tools.tokenizer.bert_tokenizer import BertTokenizer
from nlp_tools.tokenizer.jieba_tokenizer import JiebaTokenizer
from nlp_tools.tokenizer.list_tokenizer import ListTokenizer
from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer
from nlp_tools.tokenizer.whitespace_tokenizer import WhiteSpaceTokenizer
from nlp_tools.tokenizer.donothing_tokenizer import DoNothingTokenizer



from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer
from nlp_tools.utils.data import load_vocab
from typing import Dict,Any



class WhiteSpaceTokenizer(ABCTokenizer):
    """white space tokenizer
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(WhiteSpaceTokenizer, self).to_dict()
        data['config']['token_dict'] = self.token_dict
        return data

    def __init__(self, token_dict=None,max_position = 512):
        if not token_dict:
            token_dict = {}
        self.token_dict = token_dict
        # startswith = ['[PAD]', '[UNK]', '[CLS]', '[SEP]']
        if type(token_dict) != dict:
            self.token_dict = load_vocab(
                dict_path=token_dict
            )

        self.max_position = max_position
        self.id_to_token_dict = {id:key for key,id in self.token_dict.items()}

        self._token_pad_id = self.token_dict['[PAD]']
        self._token_start_id = self.token_dict['[CLS]']
        self._token_end_id = self.token_dict['[SEP]']
        self._token_mask_id = self.token_dict['[UNK]']
        self._vocab_size = len(self.token_dict)


    def tokenize(self, text, maxlen=None, **kwargs):
        if not maxlen:
            maxlen = self.max_position
        elif maxlen > self.max_position:
            maxlen = self.max_position

        if type(text) == str:
            text = text.split(" ")
            text = [t for t in text if t !=""]
            if len(text) > maxlen:
                text = text[:maxlen - 2]
        elif type(text) == list:
            if len(text) > maxlen - 2:
                text = text[:maxlen - 2]

        # 如果不以cls开头，则需要补充cls
        if text[0] != '[CLS]':
            text = ['[CLS]'] + text + ['[SEP]']
        return text


    def encode(self,
                text,
                second_text=None,
                maxlen=None,
                pattern='S*E*E',
                truncate_from='right'):
        assert type(text) in [list, str]
        if maxlen and maxlen > self.max_position:
            maxlen = self.max_position

        text_tokens = self.tokenize(text,maxlen=maxlen)
        token_ids = self.tokens_to_ids(text_tokens)
        first_segment_ids = [0] * len(token_ids)
        return token_ids, first_segment_ids
    
    def id_to_token(self, id):
        return self.id_to_token_dict[id]

    def tokens_to_ids(self,tokens):
        return [self.token_dict.get(token,1) for token in tokens]




# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for tf.keras."""

import tensorflow as tf
import tensorflow.keras.backend as K

def is_tensor_or_variable(x):
    return tf.is_tensor(x) or isinstance(x,tf.Variable)

class LossFunctionWrapper(tf.keras.losses.Loss):
    """Wraps a loss function in the `Loss` class."""
    def __init__(self,fn,reduction=tf.keras.losses.Reduction.AUTO,name=None,**kwargs):
        """Initializes `LossFunctionWrapper` class.
                Args:
                  fn: The loss function to wrap, with signature `fn(y_true, y_pred,
                    **kwargs)`.
                  reduction: (Optional) Type of `tf.keras.losses.Reduction` to apply to
                    loss. Default value is `AUTO`. `AUTO` indicates that the reduction
                    option will be determined by the usage context. For almost all cases
                    this defaults to `SUM_OVER_BATCH_SIZE`. When used with
                    `tf.distribute.Strategy`, outside of built-in training loops such as
                    `tf.keras` `compile` and `fit`, using `AUTO` or `SUM_OVER_BATCH_SIZE`
                    will raise an error. Please see this custom training [tutorial](
                      https://www.tensorflow.org/tutorials/distribute/custom_training)
                    for more details.
                  name: (Optional) name for the loss.
                  **kwargs: The keyword arguments that are passed on to `fn`.
                """
        super().__init__(reduction=reduction,name=name)
        self.fn = fn
        self._fn_kwargs = kwargs

    def call(self,y_true,y_pred):
        """Invokes the `LossFunctionWrapper` instance.
                Args:
                  y_true: Ground truth values.
                  y_pred: The predicted values.
                Returns:
                  Loss values per sample.
                """
        return self.fn(y_true,y_pred,**self._fn_kwargs)

    def get_config(self):
        config = {}
        for k,v in iter(self._fn_kwargs.items()):
            config[k] = tf.keras.backend.eval(v) if is_tensor_or_variable(v) else v
        base_config = super(LossFunctionWrapper, self).get_config().get_config()
        return {**base_config, **config}


def normalize_data_format(value):
    if value is None:
        value = tf.keras.backend.image_data_format()
    data_format = value.lower()
    if data_format not in {"channels_first", "channels_last"}:
        raise ValueError(
            "The `data_format` argument must be one of "
            '"channels_first", "channels_last". Received: ' + str(value)
        )
    return data_format


def sequence_masking(x, mask,value =0.0,axis=None):
    """为序列条件mask的函数
        mask: 形如(batch_size, seq_len)的0-1矩阵；
        value: mask部分要被替换成的值，可以是'-inf'或'inf'；
        axis: 序列所在轴，默认为1；
        """
    if mask is None:
        return x
    else:
        if K.dtype(mask) != K.dtype(x):
            mask = K.cast(mask,K.dtype(x))

        if value == '-inf':
            value = -1e12
        elif value == 'inf':
            value = 1e12

        if axis is None:
            axis =1
        elif axis < 0:
            axis = K.ndim(x) + axis

        assert  axis > 0, 'axis must be greater than 0'
        for _ in range(axis - 1):
            mask = K.expand_dims(mask, 1)

        for _ in range(K.ndim(x) - K.ndim(mask)):
            mask = K.expand_dims(mask,K.ndim(mask))

        return x * mask + value * (1 - mask)


from typing import List,Any,Dict
import tensorflow as tf
from tensorflow import keras

from nlp_tools.tasks.abs_task_model import ABCTaskModel


class EvalCallBack(keras.callbacks.Callback):

    def __init__(self,
                 model:ABCTaskModel,
                 x_data: List[Any],
                 y_data: List[Any],
                 *,
                 step: int = 5,
                 truncating: bool = False,
                 batch_size: int = 64) -> None:
        """
        Evaluate callback, calculate precision, recall and f1
        Args:
            kash_model: the task model to evaluate
            x_data: feature data for evaluation
            y_data: label data for evaluation
            step: step, default 5
            truncating: truncating: remove values from sequences larger than `model.embedding.sequence_length`
            batch_size: batch size, default 64
        """
        super(EvalCallBack,self).__init__()
        self.model:ABCTaskModel = model
        self.x_data = x_data
        self.y_data = y_data
        self.step = step
        self.truncating = truncating
        self.batch_size = batch_size
        self.logs:List[Dict] = []


    def on_epoch_end(self, epoch, logs=None):
        if (epoch + 1)% self.step == 0:
            report = self.model.evaluate(
                self.x_data,
                self.y_data,
                truncating = self.truncating,
                batch_size = self.batch_size
            )

            self.logs.append(
                {
                    'precision':report['precision'],
                    'recall':report['recall'],
                    'f1-score':report['f1-score']
                }
            )

            tf.summary.scalar('eval f1-score',data=report['f1-score'],step=epoch)
            tf.summary.scalar('eval recall',data=report['recall'],step=epoch)
            tf.summary.scalar('eval precision',data=report['precision'],step=epoch)
            print(f"\nepoch: {epoch} precision: {report['precision']:.6f},"
                  f" recall: {report['recall']:.6f}, f1-score: {report['f1-score']:.6f}")




class NerF1ScoreSaveCallBack(keras.callbacks.Callback):

    def __init__(self,
                 nlp_tools_model,
                 x_data: List[Any],
                 y_data: List[Any],
                 model_save_path=None,
                 truncating: bool = False,
                 batch_size: int = 64) -> None:
        """
        Evaluate callback, calculate precision, recall and f1
        Args:
            x_data: feature data for evaluation
            y_data: label data for evaluation
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            batch_size: batch size, default 64
        """
        super(NerF1ScoreSaveCallBack,self).__init__()
        self.nlp_tools_model = nlp_tools_model
        self.x_data = x_data
        self.y_data = y_data
        self.truncating = truncating
        self.batch_size = batch_size
        self.model_save_path = model_save_path
        self.best_f1_score = 0


    def on_epoch_end(self, epoch, logs=None):
        report = self.nlp_tools_model.evaluate(
            self.x_data,
            self.y_data,
            truncating=self.truncating,
            batch_size = self.batch_size,

        )
        f1_score = report['f1-score']

        print("current f1_score: %s,best f1_score:%s" % (f1_score,self.best_f1_score))
        if f1_score >= self.best_f1_score:
            if self.model_save_path:
                self.nlp_tools_model.save(self.model_save_path)
            self.best_f1_score = f1_score





from typing import Dict,Any
from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.simalary.abc_model import ABCSimilaryModel
import tensorflow as tf
from nlp_tools.layers.kmax_pool_layer import KMaxPoolingLayer


class mvlstm_layer(keras.layers.Layer):
    def __init__(self,config,**kwargs):
        self.config = config
        self.layers = [
            L.Bidirectional(L.LSTM(**config['bilstm'])),

        ]
        super(mvlstm_layer, self).__init__(**kwargs)


    def call(self,x,mask=None):
        output = x
        for layer_ in self.layers:
            output = layer_(output)
        return output

class SimilaryMlpLayer(L.Layer):
    def __init__(self,config,**kwargs):
        self.config = config
        self.layers = [
            L.Dense(**config['mlp_layer']),
        ]
        super(SimilaryMlpLayer, self).__init__(**kwargs)


    def call(self,x,mask=None):
        output = x
        for layer_ in self.layers:
            output = layer_(output)
        return output

class MvLstmModel(ABCSimilaryModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            "bilstm":{
                'units':300,
                'dropout': 0.3,
                'return_sequences':True,
            },
            'kmaxpooling':{
                "k":50,
            },
            'mlp_layer': {
                "units": 128,
                'activation': 'relu'
            },
            "mlp_dropout":{
                'rate':0.3
            }
    }


    def _make_inputs(self,seq_length) -> list:
        input_left = keras.layers.Input(
            name='text_left',
            shape=(seq_length,)
        )
        input_right = keras.layers.Input(
            name='text_right',
            shape=(seq_length,)
        )
        return [input_left, input_right]

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters
        embedding_model = self.embedding.embed_model

        bilstm_layer = L.Bidirectional(L.LSTM(**config["bilstm"]))
        normal_layer = L.Lambda(lambda x:tf.math.l2_normalize(x,axis=2))
        matching_layer = L.Dot(axes=[2,2],normalize=False)
        reshape_layer = L.Reshape((-1,))
        mlp_layer = SimilaryMlpLayer(config)
        mlp_dropout = L.Dropout(**config['mlp_dropout'])
        top_k_layer = L.Lambda(lambda x:tf.nn.top_k(x, **config["kmaxpooling"], sorted=True, name=None)[0])


        # Left input and right input.
        input_pairs = self._make_inputs(self.sequence_length)
        #input_left_embed ,input_right_embed = embedding_model(input_left),embedding_model(input_right)

        embed_input_pairs = []
        for input in input_pairs:
            input = embedding_model(input)
            input = bilstm_layer(input)
            input = normal_layer(input)
            embed_input_pairs.append(input)

        matching_result = matching_layer(embed_input_pairs)
        matching_result = reshape_layer(matching_result)
        matching_result = top_k_layer(matching_result)

        mlp_result = mlp_layer(matching_result)
        mlp_result = mlp_dropout(mlp_result)

        output = self.make_output_layer(output_dim)(mlp_result)


        self.tf_model = keras.Model(input_pairs,output)









from typing import Dict, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras import layers as L

from transformers import TFRobertaPreTrainedModel, TFRobertaMainLayer, RobertaConfig, RobertaTokenizer
from transformers.file_utils import add_start_docstrings_to_model_forward, add_code_sample_docstrings
from transformers.modeling_tf_outputs import TFSequenceClassifierOutput
from transformers.modeling_tf_utils import TFSequenceClassificationLoss, get_initializer, TFModelInputType, \
    input_processing

from transformers.models.roberta.modeling_tf_roberta import ROBERTA_INPUTS_DOCSTRING,_TOKENIZER_FOR_DOC,_CHECKPOINT_FOR_DOC,_CONFIG_FOR_DOC


from nlp_tools.utils.prompt.prefix_encoder import PrefixEncoder,get_prompt_values

class TFRobertaPrefixForSequenceClassification(TFRobertaPreTrainedModel):
    _keys_to_ignore_on_load_unexpected = [r"mlm___cls", r"nsp___cls", r"cls.predictions", r"cls.seq_relationship"]
    _keys_to_ignore_on_load_missing = [r"dropout"]

    def __init__(self, config: RobertaConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        self.config = config



        self.n_layer = config.num_hidden_layers
        self.n_head = config.num_attention_heads
        self.n_embd = config.hidden_size // config.num_attention_heads
        self.pre_seq_len = 8#config.pre_seq_len
        config.pre_seq_len = 8
        config.prefix_projection = True
        config.prefix_hidden_size = 512

        self.dropout = tf.keras.layers.Dropout(rate=config.hidden_dropout_prob)
        self.prefix_tokens = np.arange(self.pre_seq_len).reshape((-1,self.pre_seq_len)).tolist()
        self.prefix_encoder = PrefixEncoder(config)

        self.roberta = TFRobertaMainLayer(config,name="roberta")


    def get_prompt(self, batch_size):
        prefix_tokens = tf.convert_to_tensor(self.prefix_tokens,dtype=tf.int32)
        prefix_tokens = tf.tile(prefix_tokens,multiples=[batch_size,1])

        past_key_values = self.prefix_encoder(prefix_tokens)
        past_key_values = tf.reshape(past_key_values,
            shape=(batch_size,self.pre_seq_len,self.n_layer * 2, self.n_head,self.n_embd))
        past_key_values = self.dropout(past_key_values)
        past_key_values = tf.transpose(past_key_values,[2, 0, 3, 1, 4])
        past_key_values = tf.split(past_key_values,self.n_layer)
        return past_key_values






    @add_start_docstrings_to_model_forward(ROBERTA_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    @add_code_sample_docstrings(
        processor_class=_TOKENIZER_FOR_DOC,
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=TFSequenceClassifierOutput,
        config_class=_CONFIG_FOR_DOC,
    )
    def call(
            self,
            input_ids: Optional[TFModelInputType] = None,
            attention_mask: Optional[Union[np.ndarray, tf.Tensor]] = None,
            token_type_ids: Optional[Union[np.ndarray, tf.Tensor]] = None,
            position_ids: Optional[Union[np.ndarray, tf.Tensor]] = None,
            head_mask: Optional[Union[np.ndarray, tf.Tensor]] = None,
            inputs_embeds: Optional[Union[np.ndarray, tf.Tensor]] = None,
            output_attentions: Optional[bool] = None,
            output_hidden_states: Optional[bool] = None,
            return_dict: Optional[bool] = None,
            training: Optional[bool] = False,
            **kwargs,
    ) -> Union[TFSequenceClassifierOutput, Tuple[tf.Tensor]]:
        r"""
                labels (:obj:`tf.Tensor` or :obj:`np.ndarray` of shape :obj:`(batch_size,)`, `optional`):
                    Labels for computing the sequence classification/regression loss. Indices should be in :obj:`[0, ...,
                    config.num_labels - 1]`. If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
                    If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).
                """
        batch_size = tf.shape(input_ids['input_ids'])[0] # 4
        past_key_values = self.get_prompt(batch_size=batch_size) # shape=[2,4,16,8,64]
        prefix_attention_mask = tf.ones((batch_size, self.pre_seq_len),dtype=tf.int32)
        attention_mask = L.concatenate((prefix_attention_mask, attention_mask), axis=1) # shape [4,136]


        inputs = input_processing(
            func=self.call,
            config=self.config,
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            training=training,
            kwargs_call=kwargs,
            past_key_values=past_key_values,
        )








        outputs = self.roberta(
            input_ids=inputs["input_ids"],
            attention_mask=attention_mask,
            token_type_ids=inputs["token_type_ids"],
            position_ids=inputs["position_ids"],
            head_mask=inputs["head_mask"],
            inputs_embeds=inputs["inputs_embeds"],
            output_attentions=inputs["output_attentions"],
            output_hidden_states=inputs["output_hidden_states"],
            return_dict=inputs["return_dict"],
            #training=inputs["training"],
            past_key_values=inputs["past_key_values"],
        )

        pooled_output = outputs[1]
        pooled_output = self.dropout(inputs=pooled_output, training=inputs["training"])
        logits = self.classifier(inputs=pooled_output)
        loss = None if inputs["labels"] is None else self.compute_loss(labels=inputs["labels"], logits=logits)
        if not inputs["return_dict"]:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TFSequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



if __name__ == '__main__':



    model_name = "roberta-large"

    config = RobertaConfig.from_pretrained(model_name)
    model = TFRobertaPrefixForSequenceClassification(config)
    from transformers import BertTokenizer
    tokenizer = RobertaTokenizer.from_pretrained(model_name)
    tf_batch = tokenizer(
        ['we are very happy to show you the transformers library.',
         "we hope you don't hate it.",
         ],
        padding=True,
        truncation=True,
        max_length=128,
        return_tensors='tf',
    )
    #tf_outputs = model(tf_batch)
    tf_outputs = model(tf_batch, attention_mask = tf_batch['attention_mask'],labels=tf.constant([1, 0]))
    print(tf_outputs)


import nlp_tools
import tensorflow as tf
from tensorflow.keras.layers import Layer,concatenate,Lambda
import tensorflow.keras.backend as K
def compute_euclidean_distance(x, y):
    """
    Computes the euclidean distance between two tensorflow variables
    """

    d = tf.reduce_sum(tf.square(tf.subtract(x, y)),1)
    return d

def cosine_similarity(self, x1, x2):
    """Compute cosine similarity.
    # Arguments:
        x1: (..., embedding_size)
        x2: (..., embedding_size)
    """
    cos = tf.reduce_sum(x1 * x2, axis=-1)
    x1_norm = tf.sqrt(tf.maximum(tf.reduce_sum(tf.square(x1), axis=-1), self.epsilon))
    x2_norm = tf.sqrt(tf.maximum(tf.reduce_sum(tf.square(x2), axis=-1), self.epsilon))
    cos = cos / x1_norm / x2_norm
    return cos


class MatchingLayer(Layer):
    """
        Layer that computes a matching matrix between samples in two tensors.
        :param normalize: Whether to L2-normalize samples along the
            dot product axis before taking the dot product.
            If set to True, then the output of the dot product
            is the cosine proximity between the two samples.
        :param matching_type: the similarity function for matching
        :param kwargs: Standard layer keyword arguments.
        """
    def __init__(self,normalize: bool= False,
                 matching_type:str='dot',**kwargs):
        super().__init__(**kwargs)
        self._noramlize = normalize
        self._validate_matching_type(matching_type)
        self._matching_type = matching_type
        self._shape1 = None
        self._shape2 = None

    @classmethod
    def _validate_matching_type(cls, matching_type: str = 'dot'):
        valid_matching_type = ['dot', 'mul', 'plus', 'minus', 'concat']
        if matching_type not in valid_matching_type:
            raise ValueError(f"{matching_type} is not a valid matching type, "
                             f"{valid_matching_type} expected.")

    def build(self,input_shape:list):
        """
                Build the layer.
                :param input_shape: the shapes of the input tensors,
                    for MatchingLayer we need tow input tensors.
                """
        # Used purely for shape validation.
        if not isinstance(input_shape,list) or len(input_shape) != 2:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on a list of 2 inputs.')

        self._shape1 = input_shape[0]
        self._shape2 = input_shape[1]
        for idx in 0,2:
            if self._shape1[idx] != self._shape2[idx]:
                raise ValueError(
                    'Incompatible dimensions: '
                    f'{self._shape1[idx]} != {self._shape2[idx]}.'
                    f'Layer shapes: {self._shape1}, {self._shape2}.'
                )

    def call(self,inputs:list,**kwargs):
        """
                The computation logic of MatchingLayer.
                :param inputs: two input tensors.
                """
        x1 = inputs[0]
        x2 = inputs[1]
        if self._matching_type == 'dot':
            if self._noramlize:
                x1 = tf.math.l2_normalize(x1,axis=2)
                x2 = tf.math.l2_normalize(x2,axis=2)
            return tf.expand_dims(tf.einsum('abd,acd->abc',x1,x2),3)
        else:
            if self._matching_type == 'mul':
                def func(x,y):
                    return  x*y
            elif self._matching_type == 'plus':
                def func(x,y):
                    return x +y
            elif self._matching_type == 'minus':
                def func(x,y):
                    return x - y
            elif self._matching_type == 'concat':
                def func(x,y):
                    return tf.concat([x,y],axis=3)
            else:
                raise ValueError(f"Invalid matching type."
                                 f"{self._matching_type} received."
                                 f"Mut be in `dot`, `mul`, `plus`, "
                                 f"`minus` and `concat`.")
            x1_exp = tf.stack([x1] * self._shape2[1],2)
            x2_exp = tf.stack([x2] * self._shape1[1],1)
            return func(x1_exp,x2_exp)

    def compute_output_shape(self, input_shape):
        """
                Calculate the layer output shape.
                :param input_shape: the shapes of the input tensors,
                    for MatchingLayer we need tow input tensors.
                """
        if not isinstance(input_shape,list) or len(input_shape) != 2:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on a list of 2 inputs.')
        shape1 = list(input_shape[0])
        shape2 = list(input_shape[1])
        if len(shape1) != 3 or len(shape2) != 3:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on 2 inputs with 3 dimensions.')
        if shape1[0] != shape2[0] or shape1[2] != shape2[2]:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on 2 inputs with same 0,2 dimensions.')

        if self._matching_type in ['mul','plus','minus']:
            return shape1[0],shape1[1],shape2[1],shape1[2]
        elif self._matching_type == 'dot':
            return shape1[0],shape1[1],shape2[1],1
        elif self._matching_type == 'concat':
            return shape1[0],shape1[1],shape2[1],shape1[2]+shape2[2]
        else:
            raise ValueError(f"Invalid `matching_type`."
                             f"{self._matching_type} received."
                             f"Must be in `mul`, `plus`, `minus` "
                             f"`dot` and `concat`.")
    def get_config(self)->dict:
        config = {
            'normalize':self._noramlize,
            'matching_type':self._matching_type
        }
        base_config = super(MatchingLayer,self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class LocalInferenceLayer(object):
    """
    Layer to compute local inference between two encoded sentences a and b.
    """

    def __call__(self, inputs):
        a = inputs[0]
        b = inputs[1]

        attention = Lambda(self._attention,
                           self._attention_output_shape)(inputs)

        align_b = Lambda(self._soft_alignment,
                         self._soft_alignment_output_shape)([attention, a])
        attention = K.permute_dimensions(attention, (0, 2, 1))
        align_a = Lambda(self._soft_alignment,
                         self._soft_alignment_output_shape)([attention, b])

        # Enhancement of the local inference information obtained with the
        # attention mecanism and soft alignments.
        sub_a_align = Lambda(lambda x: x[0]-x[1])([a, align_a])
        sub_b_align = Lambda(lambda x: x[0]-x[1])([b, align_b])

        mul_a_align = Lambda(lambda x: x[0]*x[1])([a, align_a])
        mul_b_align = Lambda(lambda x: x[0]*x[1])([b, align_b])

        m_a = concatenate([a, align_a, sub_a_align, mul_a_align])
        m_b = concatenate([b, align_b, sub_b_align, mul_b_align])

        return m_a, m_b

    def _attention(self, inputs):
        """
        Compute the attention between elements of two sentences with the dot
        product.
        Args:
            inputs: A list containing two elements, one for the first sentence
                    and one for the second, both encoded by a BiLSTM.
        Returns:
            A tensor containing the dot product (attention weights between the
            elements of the two sentences).
        """
        attn_weights = K.batch_dot(x=inputs[0],
                                   y=K.permute_dimensions(inputs[1],
                                                          pattern=(0, 2, 1)))
        return K.permute_dimensions(attn_weights, (0, 2, 1))

    def _attention_output_shape(self, inputs):
        input_shape = inputs[0]
        embedding_size = input_shape[1]
        return (input_shape[0], embedding_size, embedding_size)

    def _soft_alignment(self, inputs):
        """
        Compute the soft alignment between the elements of two sentences.
        Args:
            inputs: A list of two elements, the first is a tensor of attention
                    weights, the second is the encoded sentence on which to
                    compute the alignments.
        Returns:
            A tensor containing the alignments.
        """
        attention = inputs[0]
        sentence = inputs[1]

        # Subtract the max. from the attention weights to avoid overflows.
        exp = K.exp(attention - K.max(attention, axis=-1, keepdims=True))
        exp_sum = K.sum(exp, axis=-1, keepdims=True)
        softmax = exp / exp_sum

        return K.batch_dot(softmax, sentence)

    def _soft_alignment_output_shape(self, inputs):
        attention_shape = inputs[0]
        sentence_shape = inputs[1]
        return (attention_shape[0], attention_shape[1], sentence_shape[2])


class MultiPerspective(Layer):
    """Multi-perspective Matching Layer.
    # Arguments
        mp_dim: single forward/backward multi-perspective dimention
    """

    def __init__(self, mp_dim, epsilon=1e-6, **kwargs):
        self.mp_dim = mp_dim
        self.epsilon = 1e-6
        self.strategy = 4
        super(MultiPerspective, self).__init__(**kwargs)

    def build(self, input_shape):
        if isinstance(input_shape, list):
            input_shape = input_shape[0]
        embedding_size = int(input_shape[-1] / 2)
        # Create a trainable weight variable for this layer.
        # input_shape is bidirectional RNN input shape
        # kernel shape (mp_dim * 2 * self.strategy, embedding_size)
        self.kernel = self.add_weight(name='kernel',
                                      shape=(self.mp_dim,
                                             embedding_size * 2 * self.strategy),
                                      initializer='glorot_uniform',
                                      trainable=True)
        self.kernel_full_fw = self.kernel[:, :embedding_size]
        self.kernel_full_bw = self.kernel[:, embedding_size: embedding_size * 2]
        self.kernel_attentive_fw = self.kernel[:, embedding_size * 2: embedding_size * 3]
        self.kernel_attentive_bw = self.kernel[:, embedding_size * 3: embedding_size * 4]
        self.kernel_max_attentive_fw = self.kernel[:, embedding_size * 4: embedding_size * 5]
        self.kernel_max_attentive_bw = self.kernel[:, embedding_size * 5: embedding_size * 6]
        self.kernel_max_pool_fw = self.kernel[:, embedding_size * 6: embedding_size * 7]
        self.kernel_max_pool_bw = self.kernel[:, embedding_size * 7:]
        self.built = True
        super(MultiPerspective, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        if isinstance(input_shape, list):
            input_shape = input_shape[0]
        return (input_shape[0], input_shape[1], self.mp_dim * 2 * self.strategy)

    def get_config(self):
        config = {'mp_dim': self.mp_dim,
                  'epsilon': self.epsilon}
        base_config = super(MultiPerspective, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def call(self, inputs):
        # h1, h2: bidirectional LSTM hidden states, include forward and backward states
        #         (batch_size, timesteps, embedding_size * 2)
        h1 = inputs[0]
        h2 = inputs[1]
        embedding_size = int(K.int_shape(h1)[-1] / 2)
        h1_fw = h1[:, :, :embedding_size]
        h1_bw = h1[:, :, embedding_size:]
        h2_fw = h2[:, :, :embedding_size]
        h2_bw = h2[:, :, embedding_size:]

        # 4 matching strategy
        list_matching = []

        # full matching ops
        matching_fw = self._full_matching(h1_fw, h2_fw, self.kernel_full_fw)
        matching_bw = self._full_matching(h1_bw, h2_bw, self.kernel_full_bw)
        list_matching.extend([matching_fw, matching_bw])

        # cosine matrix
        cosine_matrix_fw = self._cosine_matrix(h1_fw, h2_fw)
        cosine_matrix_bw = self._cosine_matrix(h1_bw, h2_bw)

        # attentive matching ops
        matching_fw = self._attentive_matching(
            h1_fw, h2_fw, cosine_matrix_fw, self.kernel_attentive_fw)
        matching_bw = self._attentive_matching(
            h1_bw, h2_bw, cosine_matrix_bw, self.kernel_attentive_bw)
        list_matching.extend([matching_fw, matching_bw])

        # max attentive matching ops
        matching_fw = self._max_attentive_matching(
            h1_fw, h2_fw, cosine_matrix_fw, self.kernel_max_attentive_fw)
        matching_bw = self._max_attentive_matching(
            h1_bw, h2_bw, cosine_matrix_bw, self.kernel_max_attentive_bw)
        list_matching.extend([matching_fw, matching_bw])

        # max pooling matching ops
        matching_fw = self._max_pooling_matching(h1_fw, h2_fw, self.kernel_max_pool_fw)
        matching_bw = self._max_pooling_matching(h1_bw, h2_bw, self.kernel_max_pool_bw)
        list_matching.extend([matching_fw, matching_bw])

        return K.concatenate(list_matching, axis=-1)

    def _cosine_similarity(self, x1, x2):
        """Compute cosine similarity.
        # Arguments:
            x1: (..., embedding_size)
            x2: (..., embedding_size)
        """
        cos = K.sum(x1 * x2, axis=-1)
        x1_norm = K.sqrt(K.maximum(K.sum(K.square(x1), axis=-1), self.epsilon))
        x2_norm = K.sqrt(K.maximum(K.sum(K.square(x2), axis=-1), self.epsilon))
        cos = cos / x1_norm / x2_norm
        return cos

    def _cosine_matrix(self, x1, x2):
        """Cosine similarity matrix.
        Calculate the cosine similarities between each forward (or backward)
        contextual embedding h_i_p and every forward (or backward)
        contextual embeddings of the other sentence
        # Arguments
            x1: (batch_size, x1_timesteps, embedding_size)
            x2: (batch_size, x2_timesteps, embedding_size)
        # Output shape
            (batch_size, x1_timesteps, x2_timesteps)
        """
        # expand h1 shape to (batch_size, x1_timesteps, 1, embedding_size)
        x1 = K.expand_dims(x1, axis=2)
        # expand x2 shape to (batch_size, 1, x2_timesteps, embedding_size)
        x2 = K.expand_dims(x2, axis=1)
        # cosine matrix (batch_size, h1_timesteps, h2_timesteps)
        cos_matrix = self._cosine_similarity(x1, x2)
        return cos_matrix

    def _mean_attentive_vectors(self, x2, cosine_matrix):
        """Mean attentive vectors.
        Calculate mean attentive vector for the entire sentence by weighted
        summing all the contextual embeddings of the entire sentence
        # Arguments
            x2: sequence vectors, (batch_size, x2_timesteps, embedding_size)
            cosine_matrix: cosine similarities matrix of x1 and x2,
                           (batch_size, x1_timesteps, x2_timesteps)
        # Output shape
            (batch_size, x1_timesteps, embedding_size)
        """
        # (batch_size, x1_timesteps, x2_timesteps, 1)
        expanded_cosine_matrix = K.expand_dims(cosine_matrix, axis=-1)
        # (batch_size, 1, x2_timesteps, embedding_size)
        x2 = K.expand_dims(x2, axis=1)
        # (batch_size, x1_timesteps, embedding_size)
        weighted_sum = K.sum(expanded_cosine_matrix * x2, axis=2)
        # (batch_size, x1_timesteps, 1)
        sum_cosine = K.expand_dims(K.sum(cosine_matrix, axis=-1) + self.epsilon, axis=-1)
        # (batch_size, x1_timesteps, embedding_size)
        attentive_vector = weighted_sum / sum_cosine
        return attentive_vector

    def _max_attentive_vectors(self, x2, cosine_matrix):
        """Max attentive vectors.
        Calculate max attentive vector for the entire sentence by picking
        the contextual embedding with the highest cosine similarity
        as the attentive vector.
        # Arguments
            x2: sequence vectors, (batch_size, x2_timesteps, embedding_size)
            cosine_matrix: cosine similarities matrix of x1 and x2,
                           (batch_size, x1_timesteps, x2_timesteps)
        # Output shape
            (batch_size, x1_timesteps, embedding_size)
        """
        # (batch_size, x1_timesteps)
        max_x2_step = K.argmax(cosine_matrix, axis=-1)

        embedding_size = K.int_shape(x2)[-1]
        timesteps = K.int_shape(max_x2_step)[-1]
        if timesteps is None:
            timesteps = K.shape(max_x2_step)[-1]

        # collapse time dimension and batch dimension together
        # collapse x2 to (batch_size * x2_timestep, embedding_size)
        x2 = K.reshape(x2, (-1, embedding_size))
        # collapse max_x2_step to (batch_size * h1_timesteps)
        max_x2_step = K.reshape(max_x2_step, (-1,))
        # (batch_size * x1_timesteps, embedding_size)
        max_x2 = K.gather(x2, max_x2_step)
        # reshape max_x2, (batch_size, x1_timesteps, embedding_size)
        attentive_vector = K.reshape(max_x2, K.stack([-1, timesteps, embedding_size]))
        return attentive_vector

    def _time_distributed_multiply(self, x, w):
        """Element-wise multiply vector and weights.
        # Arguments
            x: sequence of hidden states, (batch_size, ?, embedding_size)
            w: weights of one matching strategy of one direction,
               (mp_dim, embedding_size)
        # Output shape
            (?, mp_dim, embedding_size)
        """
        # dimension of vector
        n_dim = K.ndim(x)
        embedding_size = K.int_shape(x)[-1]
        timesteps = K.int_shape(x)[1]
        if timesteps is None:
            timesteps = K.shape(x)[1]

        # collapse time dimension and batch dimension together
        x = K.reshape(x, (-1, embedding_size))
        # reshape to (?, 1, embedding_size)
        x = K.expand_dims(x, axis=1)
        # reshape weights to (1, mp_dim, embedding_size)
        w = K.expand_dims(w, axis=0)
        # element-wise multiply
        x = x * w
        # reshape to original shape
        if n_dim == 3:
            x = K.reshape(x, K.stack([-1, timesteps, self.mp_dim, embedding_size]))
            x.set_shape([None, None, None, embedding_size])
        elif n_dim == 2:
            x = K.reshape(x, K.stack([-1, self.mp_dim, embedding_size]))
            x.set_shape([None, None, embedding_size])
        return x

    def _full_matching(self, h1, h2, w):
        """Full matching operation.
        # Arguments
            h1: (batch_size, h1_timesteps, embedding_size)
            h2: (batch_size, h2_timesteps, embedding_size)
            w: weights of one direction, (mp_dim, embedding_size)
        # Output shape
            (batch_size, h1_timesteps, mp_dim)
        """
        # h2 forward last step hidden vector, (batch_size, embedding_size)
        h2_last_state = h2[:, -1, :]
        # h1 * weights, (batch_size, h1_timesteps, mp_dim, embedding_size)
        h1 = self._time_distributed_multiply(h1, w)
        # h2_last_state * weights, (batch_size, mp_dim, embedding_size)
        h2 = self._time_distributed_multiply(h2_last_state, w)
        # reshape to (batch_size, 1, mp_dim, embedding_size)
        h2 = K.expand_dims(h2, axis=1)
        # matching vector, (batch_size, h1_timesteps, mp_dim)
        matching = self._cosine_similarity(h1, h2)
        return matching

    def _max_pooling_matching(self, h1, h2, w):
        """Max pooling matching operation.
        # Arguments
            h1: (batch_size, h1_timesteps, embedding_size)
            h2: (batch_size, h2_timesteps, embedding_size)
            w: weights of one direction, (mp_dim, embedding_size)
        # Output shape
            (batch_size, h1_timesteps, mp_dim)
        """
        # h1 * weights, (batch_size, h1_timesteps, mp_dim, embedding_size)
        h1 = self._time_distributed_multiply(h1, w)
        # h2 * weights, (batch_size, h2_timesteps, mp_dim, embedding_size)
        h2 = self._time_distributed_multiply(h2, w)
        # reshape v1 to (batch_size, h1_timesteps, 1, mp_dim, embedding_size)
        h1 = K.expand_dims(h1, axis=2)
        # reshape v1 to (batch_size, 1, h2_timesteps, mp_dim, embedding_size)
        h2 = K.expand_dims(h2, axis=1)
        # cosine similarity, (batch_size, h1_timesteps, h2_timesteps, mp_dim)
        cos = self._cosine_similarity(h1, h2)
        # (batch_size, h1_timesteps, mp_dim)
        matching = K.max(cos, axis=2)
        return matching

    def _attentive_matching(self, h1, h2, cosine_matrix, w):
        """Attentive matching operation.
        # Arguments
            h1: (batch_size, h1_timesteps, embedding_size)
            h2: (batch_size, h2_timesteps, embedding_size)
            cosine_matrix: weights of hidden state h2,
                          (batch_size, h1_timesteps, h2_timesteps)
            w: weights of one direction, (mp_dim, embedding_size)
        # Output shape
            (batch_size, h1_timesteps, mp_dim)
        """
        # h1 * weights, (batch_size, h1_timesteps, mp_dim, embedding_size)
        h1 = self._time_distributed_multiply(h1, w)
        # attentive vector (batch_size, h1_timesteps, embedding_szie)
        attentive_vec = self._mean_attentive_vectors(h2, cosine_matrix)
        # attentive_vec * weights, (batch_size, h1_timesteps, mp_dim, embedding_size)
        attentive_vec = self._time_distributed_multiply(attentive_vec, w)
        # matching vector, (batch_size, h1_timesteps, mp_dim)
        matching = self._cosine_similarity(h1, attentive_vec)
        return matching

    def _max_attentive_matching(self, h1, h2, cosine_matrix, w):
        """Max attentive matching operation.
        # Arguments
            h1: (batch_size, h1_timesteps, embedding_size)
            h2: (batch_size, h2_timesteps, embedding_size)
            cosine_matrix: weights of hidden state h2,
                          (batch_size, h1_timesteps, h2_timesteps)
            w: weights of one direction, (mp_dim, embedding_size)
        # Output shape
            (batch_size, h1_timesteps, mp_dim)
        """
        # h1 * weights, (batch_size, h1_timesteps, mp_dim, embedding_size)
        h1 = self._time_distributed_multiply(h1, w)
        # max attentive vector (batch_size, h1_timesteps, embedding_szie)
        max_attentive_vec = self._max_attentive_vectors(h2, cosine_matrix)
        # max_attentive_vec * weights, (batch_size, h1_timesteps, mp_dim, embedding_size)
        max_attentive_vec = self._time_distributed_multiply(max_attentive_vec, w)
        # matching vector, (batch_size, h1_timesteps, mp_dim)
        matching = self._cosine_similarity(h1, max_attentive_vec)
        return matching







nlp_tools.custom_objects['MatchingLayer'] = MatchingLayer
nlp_tools.custom_objects['MultiPerspective'] = MultiPerspective








import os,json

class CLUENERGlobalPointerLoader(object):
    @classmethod
    def load_data(cls,
                  subset_name: str = 'train',
                  corpus_path=None):
        '''
        传入格式：
                {"text": "吴三桂演义》小说的想像，说是为牛金星所毒杀。……在小说中加插一些历史背景，", "label": {"book": {"吴三桂演义》": [[0, 5]]}, "name": {"牛金星": [[15, 17]]}}}
                {"text": "看来各支一二流的国家队也开始走出欧洲杯后低迷，从本期对阵情况看，似乎冷门度也不太高，你认为呢？", "label": {"organization": {"欧洲杯": [[16, 18]]}}}
        输入格式：
        data [
           ('吴三桂演义》小说的想像，说是为牛金星所毒杀。……在小说中加插一些历史背景', [   [0, 5,'book'],  [15, 17,'name'] ]),
            ('看来各支一二流的国家队也开始走出欧洲杯后低迷，从本期对阵情况看，似乎冷门度也不太高，你认为呢？',[   [16, 18,'organization']]),
        ]
        '''
        assert subset_name in ['train','test','dev']
        file_path = os.path.join(corpus_path,subset_name+".json")
        result = []
        with open(file_path, encoding='utf-8') as f:
            for line in f:
                item = []
                item_json = json.loads(line.strip())
                item.append(item_json['text'])
                sub_label = []
                for k, v in item_json['label'].items():
                    for spans in v.values():
                        for start, end in spans:
                            sub_label.append((start, end, k))
                item.append(sub_label)
                result.append(item)

        return result

if __name__ == '__main__':
    cluenew_data_path = '/home/qiufengfeng/nlp/nlp_data/ner/cluener'
    print(CLUENERGlobalPointerLoader.load_data(corpus_path=cluenew_data_path))



import json
from typing import Dict, List, Any
import numpy as np
import tensorflow as tf
import nlp_tools


L = tf.keras.layers


class ABCEmbedding:
    def to_dict(self) -> Dict[str, Any]:
        config: Dict[str, Any] = {
            'segment': self.segment,
            'embedding_size': self.embedding_size,
            'max_position': self.max_position,
            'bert_application':self.bert_application,
            **self.kwargs
        }
        return {
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'config': config,
            #'embed_model': json.loads(self.embed_model.to_json())
        }

    def __init__(self,
                 segment: bool = False,
                 embedding_size: int = 100,
                 max_position: int = None,
                 bert_application='encoder',
                 **kwargs: Any):

        self.embed_model: tf.keras.Model = None

        self.segment: bool = segment  # type: ignore
        self.kwargs = kwargs

        self.embedding_size: int = embedding_size  # type: ignore
        self.max_position: int = max_position  # type: ignore
        self.bert_application = bert_application

    def _override_load_model(self,config:Dict) -> None:
        pass
        # embed_model_json_str = json.dumps(config['embed_model'])
        # self.embed_model = tf.keras.models.model_from_json(embed_model_json_str,
        #                                                    custom_objects=nlp_tools.custom_objects)


    def build_embedding_model(self,
                              *,
                              vocab_size: int = None,
                              force: bool = False,
                              bert_application=None,
                              **kwargs: Dict) -> None:
        raise NotImplementedError

    def embed(self,tensor_x: List[List[List[int]]],) -> np.ndarray:
        """
        batch embed sentences

        Args:
            sentences: Sentence list to embed
            debug: show debug info
        Returns:
            vectorized sentence list
        """

        embed_results = self.embed_model.predict(tensor_x)
        return embed_results




from typing import Dict,Any
from tensorflow import keras

from nlp_tools.tasks.scoring.base_model import BaseScoringModel
from nlp_tools.layers import L

class BiLSTM_Model(BaseScoringModel):
    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str,Dict[str,Any]]:
        return {
            'layer_bi_lstm':{
                'units':128,
                'return_sequences':False
            },
            'layer_dense':{
                'activation':"linear"
            }
        }

    def build_model_arc(self):
        output_dim = self.processor.output_dim
        config = self.hyper_parameters
        embed_model =self.embedding.embed_model

        layer_bi_lstm = L.Bidirectional(L.LSTM(**config['layer_bi_lstm']))
        layer_dense = L.Dense(output_dim,**config['layer_dense'])

        tensor = layer_bi_lstm(embed_model.output)
        output_tensor = layer_dense(tensor)

        self.tf_model = keras.Model(embed_model.inputs, output_tensor)


if __name__ == "__main__":
    from nlp_tools.corpus import SMP2018ECDTCorpus
    import numpy as np

    x, y = SMP2018ECDTCorpus.load_data('valid')
    y = np.random.random((len(x), 4))
    model = BiLSTM_Model()
    model.fit(x, y)
    print(model.predict(x[:10]))


from typing import Dict,Any
from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.simalary.abc_model import ABCSimilaryModel
from tensorflow.keras.losses import cosine_similarity

class dssm_layer(keras.layers.Layer):
    def __init__(self,config,**kwargs):
        self.config = config
        self.layers = [
            L.Conv1D(**self.config['conv1']),
            L.Dropout(**self.config['dropout']),
            L.GlobalMaxPool1D(**self.config['global_pool']),
            L.Dense(**self.config['mlp_layer']),
            L.Dense(**self.config['out_layer']),
        ]
        super(dssm_layer, self).__init__(**kwargs)


    def call(self,x,mask=None):
        output = x
        for layer_ in self.layers:
            output = layer_(output)
        return output


class CDssmModel(ABCSimilaryModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            "conv1":{
                'filters':300,
                'kernel_size': 3,
                'strides': 1,
                'padding': 'same',
                'activation':'relu',
                "kernel_initializer":"random_uniform",
                "bias_initializer":"zeros"
            },
            'dropout':{
                'rate':0.2
            },
            'global_pool':{

            },
            'mlp_layer':{
                "units":128,
                'activation':'relu'
            },
            'out_layer':{
                "units": 128,
                'activation': 'relu'
            }
    }

    def _create_base_network(self):
        def _wrapper(x):
            embedding_model = self.embedding.embed_model
            config = self.default_hyper_parameters()

            layers = [
                embedding_model,
                L.Conv1D(**config['conv1']),
                L.Dropout(**config['dropout']),
                L.GlobalMaxPool1D(**config['global_pool']),
                L.Dense(**config['mlp_layer']),
                L.Dense(**config['out_layer']),
            ]
            output  = x
            for layer_ in layers:
                output = layer_(output)
            return output
        return _wrapper





    def base_encode_block(self,config):
        layer =  L.Bidirectional(L.LSTM(**config['layer_bi_lstm']))
        return layer

    def _make_inputs(self,seq_length) -> list:
        input_left = keras.layers.Input(
            name='text_left',
            shape=(seq_length,)
        )
        input_right = keras.layers.Input(
            name='text_right',
            shape=(seq_length,)
        )
        return [input_left, input_right]

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters
        embedding_model = self.embedding.embed_model
        base_network = dssm_layer(config)


        # Left input and right input.
        input_left, input_right = self._make_inputs(self.sequence_length)
        input_left_embed ,input_right_embed = embedding_model(input_left),embedding_model(input_right)
        # Process left & right input.
        x = [base_network(input_left_embed),
             base_network(input_right_embed)]

        # Dot product with cosine similarity.
        x = L.Dot(axes=[1, 1], normalize=True)(x)

        output = L.Dense(output_dim, activation='softmax')(x)

        self.tf_model = keras.Model([input_left,input_right],output)


    def compile_model(self,
                      loss:Any = None,
                      optimizer: Any=None,
                      metrics:Any=None,
                      **kwargs:Any) -> None:
        if loss is None:
            loss = 'sparse_categorical_crossentropy'

        if optimizer is None:
            optimizer = 'adam'

        if metrics is None:
            metrics=["cosine_similarity"]

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)






from tensorflow.keras.layers import Layer
from tensorflow.keras.initializers import TruncatedNormal
import tensorflow.keras.backend  as K
import tensorflow as tf



class NormalAttentionLayer(Layer):
    def __init__(self,attention_dim,**kwargs):
        self.init = TruncatedNormal()
        self.supports_masking = True
        self.attention_dim = attention_dim
        super(NormalAttentionLayer, self).__init__(**kwargs)


    def build(self, input_shape):
        #assert len(input_shape) == 3
        self.W = self.add_weight(name='W',
                                 shape=(input_shape[-1],self.attention_dim),
                                 dtype=K.floatx(),
                                 initializer=self.init,
                                 trainable=True
                                 )

        self.b = self.add_weight(name='b',
                                 shape=(self.attention_dim,),
                                 dtype=K.floatx(),
                                 initializer=self.init,
                                 trainable=True
                                 )

        self.u = self.add_weight(name='u',
                                 shape=(self.attention_dim,1),
                                 dtype=K.floatx(),
                                 initializer=self.init,
                                 trainable=True
                                 )
        super(NormalAttentionLayer, self).build(input_shape)

    def compute_mask(self, inputs, mask=None):
        if mask != None:
            mask = K.cast(mask, tf.int32)
            mask_sum = K.sum(mask, axis=-1)
            mask_new = K.cast(mask_sum > 0, tf.bool)  # sentence mask
            return mask_new
        else:
            return mask

    def call(self,x,mask=None):
        # size of x :[batch_size, sel_len, attention_dim]
        # size of u :[batch_size, attention_dim]
        uit = K.tanh(K.bias_add(K.dot(x,self.W),self.b))
        ait = K.dot(uit,self.u)
        ait = K.squeeze(ait,-1)

        ait = K.exp(ait)

        if mask is not None:
            ait *= K.cast(mask,K.floatx())

        ait /= K.cast(K.sum(ait,axis=-2,keepdims=True) + K.epsilon(), K.floatx())
        ait = K.expand_dims(ait)

        weighted_input = x * ait
        output = K.sum(weighted_input, axis=1)

        return output

    def compute_output_shape(self, input_shape):
        return (input_shape[0],input_shape[-1])

    def get_config(self):
        config = {
            'attention_dim':self.attention_dim
        }
        base_config = super(NormalAttentionLayer,self).get_config()
        return dict(list(base_config.items()) + list(config.items()))






import collections
import operator
from typing import List, Union, Dict, Optional, Any, Tuple

import numpy as np
import tqdm

from nlp_tools.generators import CorpusGenerator
from nlp_tools.processors.abc_label_processor import ABCLabelProcessor
from nlp_tools.types import  TextSamplesVar
from nlp_tools.utils import MultiLabelBinarizer
from tensorflow.keras.utils import to_categorical

class ClassificationLabelProcessor(ABCLabelProcessor):

    def to_dict(self) -> Dict[str, Any]:
        data = super(ClassificationLabelProcessor, self).to_dict()
        return data

    def __init__(self,label_smooth=False,
                 **kwargs: Any) -> None:
        self.label_smooth = label_smooth
        super(ClassificationLabelProcessor, self).__init__(**kwargs)



    def build_vocab_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        vocab2idx: Dict[str, int] = {}
        token2count: Dict[str, int] = {}
        for generator in generators:
            for _, label in tqdm.tqdm(generator.sample(), desc="Preparing classification label vocab dict"):
                count = token2count.get(label, 0)
                token2count[label] = count + 1

        sorted_token2count = sorted(token2count.items(),
                                    key=operator.itemgetter(0))
        token2count = collections.OrderedDict(sorted_token2count)

        for token, token_count in token2count.items():
            if token not in vocab2idx:
                vocab2idx[token] = len(vocab2idx)
        self.vocab2idx = vocab2idx
        self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])
        self.multi_label_binarizer = MultiLabelBinarizer(self.vocab2idx)

    def get_tensor_shape(self, batch_size: int, seq_length: int) -> Tuple:
        return (batch_size,)

    def transform(self,
                  samples: TextSamplesVar,
                  *,
                  seq_length: int = None,
                  max_position: int = None,
                  segment: bool = False) -> np.ndarray:
        sample_tensor = [self.vocab2idx[i] for i in samples]

        sample_tensor = to_categorical(sample_tensor,num_classes=len(self.vocab2idx))
        if self.label_smooth:
            sample_tensor = self.smooth_labels(sample_tensor)
        return np.array(sample_tensor)

    def inverse_transform(self,
                          labels: Union[List[int], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> Union[List[List[str]], List[str]]:

        return [self.idx2vocab[i] for i in labels]

    def smooth_labels(self, labels, factor=0.2):
        # smooth the labels
        labels *= (1 - factor)
        labels += (factor / labels.shape[1])

        return labels





# encoding: utf-8


import collections
import operator
from typing import Dict, List, Any, Optional, Union

import numpy as np
import tqdm
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.processors.abc_sentence_processor import ABCSentenceProcessor
from nlp_tools.types import TextSamplesVar


class DialogueSequenceProcessor(ABCSentenceProcessor):
    """
    Generic processors for the sequence samples.
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(DialogueSequenceProcessor, self).to_dict()
        data['config'].update({
        })
        return data

    def __init__(self,
                 return_ner_masking=True,
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        self.return_ner_masking = return_ner_masking
        super(DialogueSequenceProcessor, self).__init__(**kwargs)



    def update_length_info(self,embedding_max_position=512,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length

    def transform(self,samples: TextSamplesVar,seq_length: int = None) :

        if seq_length is None and self.max_sentence_length is not None:
            seq_length = self.max_sentence_length/2

        if seq_length is None:
            seq_length = int(self.embedding_max_position/4)
        elif self.embedding_max_position < seq_length:
            seq_length = self.embedding_max_position




        ner_masking = []
        tokenids = []
        segment_ids = []

        for seq in samples:
            sub_tokens_id = []
            sub_segment_id = []
            sub_ner_masking = []
            start_position = 0
            for index, sub_seqn in enumerate(seq):
                if "$$"  in sub_seqn:
                    split_list = sub_seqn.split("$$")
                    sub_seqn = "".join(split_list)

                    start_entity = start_position+len(split_list[0])
                    end_entity = start_position+len(split_list[0])+len(split_list[1])
                    for i in range(start_entity,end_entity):
                        sub_ner_masking.append(i)
                    # sub_ner_masking.append(start_position+len(split_list[0]))
                    # sub_ner_masking.append(start_position+len(split_list[0])+len(split_list[1])-1)
                seq_tokenids,seq_segment_ids = self.text_tokenizer.encode(sub_seqn,maxlen=seq_length)
                if index == 0:
                    sub_tokens_id.extend(seq_tokenids)
                    sub_segment_id.extend(seq_segment_ids)
                    start_position += len(seq_tokenids)
                else:
                    sub_tokens_id.extend(seq_tokenids[1:])
                    sub_segment_id.extend(seq_segment_ids[1:])
                    start_position += len(seq_tokenids) - 1

            tokenids.append(sub_tokens_id)
            segment_ids.append(sub_segment_id)
            ner_masking.append(sub_ner_masking)



        token_ids = pad_sequences(tokenids, seq_length, padding='post', truncating='post')
        segment_ids = pad_sequences(segment_ids, seq_length, padding='post', truncating='post')
        ner_masking_ids = pad_sequences(ner_masking, max([len(i) for i in ner_masking]), padding='post', truncating='post')
        token_ids = np.array(token_ids)
        segment_ids = np.array(segment_ids)
        ner_masking_ids = np.array(ner_masking_ids)

        return token_ids,segment_ids,ner_masking_ids




    def inverse_transform(self,
                          labels: Union[List[List[int]], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> List[List[str]]:
        result = []
        for index, seq in enumerate(labels):
            labels_ = []
            for idx in seq:
                labels_.append(self.text_tokenizer.id_to_token(idx))
            if lengths is not None:
                labels_ = labels_[1:lengths[index] + 1]
            else:
                labels_ = labels_[1:-1]
            result.append(labels_)
        return result

    def get_max_seq_length_from_corpus(self,
                                   generators: List[CorpusGenerator],
                                   cover_rate: float = 0.95) -> int:
        """
        Calculate proper sequence length according to the corpus

        Args:
            generators:
            use_label:
            cover_rate:

        Returns:

        """
        return 512



if __name__ == "__main__":
    pass



from tensorflow.keras import backend as K
from tensorflow.python.keras.metrics import MeanMetricWrapper
import tensorflow as tf

@tf.keras.utils.register_keras_serializable(package='nlp_tools',name='GlobalPointerF1Score')
def global_pointer_f1_score(y_true, y_pred):
    """给GlobalPointer设计的F1
    """
    y_pred = K.cast(K.greater(y_pred, 0), K.floatx())
    return 2 * K.sum(y_true * y_pred) / K.sum(y_true + y_pred)

# @tf.keras.utils.register_keras_serializable(package='nlp_tools',name='GlobalPointerF1Score')
# class GlobalPointerF1Score():
#     def __init__(self, name='global_pointer_f1_score', dtype=None):
#         super(GlobalPointerF1Score, self).__init__(global_pointer_f1_score, name, dtype=dtype)




from tensorflow.keras import Model,Sequential
from tensorflow.keras.layers import Embedding,Dense
import tensorflow as tf
import numpy as np
from typing import Union

from nlp_tools.embeddings.hugginface.prefix_prompt.configuration_prefix_prompt_roberta import PrefixPromptConfig
from transformers import BertConfig

class PrefixEncoder(Model):
    r'''
        The tensorflow model to encode the prefix

        Input shape: (batch-size, prefix-length)

        Output shape: (batch-size, prefix-length, 2*layers*hidden)
        '''
    def __init__(self, config:Union[PrefixPromptConfig,BertConfig]):
        super().__init__()
        self.prefix_projection = config.prefix_projection
        if self.prefix_projection:
            # Use a two-layer MLP to encode the prefix
            self.embedding = Embedding(config.pre_seq_len, config.hidden_size)
            self.trans = Sequential([
                Dense(config.prefix_hidden_size,activation='tanh'),
                Dense(config.num_hidden_layers * 2 * config.hidden_size)]
            )
        else:
            self.embedding = Embedding(config.pre_seq_len, config.num_hidden_layers * 2 * config.hidden_size)

    def call(self,prefix,**kwargs):
        if self.prefix_projection:
            prefix_tokens = self.embedding(prefix)
            past_key_values = self.trans(prefix_tokens)
        else:
            past_key_values = self.embedding(prefix)
        return past_key_values


def get_prompt_values(config:Union[PrefixPromptConfig,BertConfig],batch_size):
    '''
    生成需要拼接到
    :param config: bert模型或者其他模型的config对象实例
    :param batch_size: 传入的训练数据的大小，
    :param pre_seq_len: prefix的输入字符的长度
    :param bert_num_hidden_layers: 需要添加到的bert或者其他模型的总层数量（bert里面相当于num_hidden_layers）
    :param bert_num_attention_heads: bert里面多头注意力的个数
    :param bert_hidden_size: bert里面隐层的unit大小（bert里面是768）
    :return: prefix_values_tensor,prefix_attenion_mask
    '''
    n_embd = config.hidden_size//config.num_attention_heads

    prefix_tokens = np.arange(config.pre_seq_len).reshape((-1, config.pre_seq_len)).tolist()
    prefix_tokens = tf.convert_to_tensor(prefix_tokens,dtype=tf.int32)
    prefix_tokens = tf.tile(prefix_tokens,multiples=[batch_size,1])


    # 定义相应的模型层
    prefix_encoder_layer = PrefixEncoder(config)
    dropout_layer = tf.keras.layers.Dropout(rate=config.hidden_dropout_prob)

    past_key_values = prefix_encoder_layer(prefix_tokens)
    past_key_values = tf.reshape(past_key_values,
        shape=(batch_size,config.pre_seq_len,config.num_hidden_layers * 2, config.num_attention_heads,n_embd))
    past_key_values = dropout_layer(past_key_values)
    past_key_values = tf.transpose(past_key_values,[2, 0, 3, 1, 4])
    past_key_values = tf.split(past_key_values,config.num_hidden_layers)

    prefix_attention_mask = tf.ones((batch_size, config.pre_seq_len), dtype=tf.int32)
    return past_key_values,prefix_attention_mask



import random
import os
from abc import ABC
import numpy as np
from typing import List,Dict,Any,Union
from tensorflow import keras

import nlp_tools
from nlp_tools.embeddings import BertEmbedding
from nlp_tools.generators import BertGeneratorDataSet,CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.tasks.abs_task_model import ABCTaskModel
from nlp_tools.types import TextSamplesVar,ClassificationLabelVar,MultiLabelClassificationLabelVar
from nlp_tools.loss import TextGenerateCrossEntropy
from nlp_tools.tokenizer.bert_tokenizer import BertTokenizer
from nlp_tools.processors.sequence_processor import SequenceProcessor
from tensorflow.keras.optimizers import Adam



class BertGeneratorModel(ABCTaskModel,ABC):
    """
    Abstract Classification Model
    """

    __task__ = 'generator'

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            }

    def to_dict(self) -> Dict:
        info = super(BertGeneratorModel, self).to_dict()
        info['config']['bert_model_path'] = self.bert_model_path
        info['config']['sequence_length'] = self.sequence_length
        info['config']['hyper_parameters'] = self.hyper_parameters
        info['config']['bert_application'] = self.bert_application
        info['config']['bert_token_simplified'] = self.bert_token_simplified
        return info

    def __init__(self,
                 bert_model_path ,
                 text_processor = None,
                 sequence_length: int = None,
                 hyper_parameters: Dict[str, Dict[str, Any]] = None,
                 bert_application='unilm',
                 bert_token_simplified=True):
        """
        Args:
            sequence_length: target sequence length
            hyper_parameters: hyper_parameters to overwrite
            text_processor: text processor
        """
        super(BertGeneratorModel, self).__init__()

        if text_processor == None:

            bert_vocab_path = os.path.join(bert_model_path,'vocab.txt')
            text_tokenizer = BertTokenizer(token_dict=bert_vocab_path,simplified=bert_token_simplified)
            text_processor = SequenceProcessor(text_tokenizer=text_tokenizer)

        if type(text_processor.text_tokenizer) != BertTokenizer:
            raise("bert Embedding must use bertTokenizer")

        self.embedding = BertEmbedding(model_folder=bert_model_path,
                                       bert_application=bert_application,
                                       keep_tokens=text_processor.text_tokenizer.keep_tokens,
                                       simplified=text_processor.text_tokenizer.simplified)  # type: ignore
        self.bert_model_path = bert_model_path
        self.sequence_length = sequence_length
        self.bert_application = bert_application
        self.text_processor = text_processor
        self.label_processor = text_processor
        self.hyper_parameters = hyper_parameters
        self.bert_token_simplified = bert_token_simplified


    def build_model(self,
                    x_train: TextSamplesVar,
                    y_train: TextSamplesVar) -> None:
        train_gen = CorpusGenerator(x_train, y_train)
        self.build_model_generator([train_gen])

    def build_model_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        if not self.text_processor.vocab2idx:
            self.text_processor.build_vocab_generator(generators)
        self.embedding.setup_text_processor(self.text_processor)

        if self.sequence_length is None:
            self.sequence_length = self.text_processor.get_max_seq_length_from_corpus(generators)

        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

    def build_model_arc(self) -> None:
        output = TextGenerateCrossEntropy(2)(self.embedding.embed_model.inputs + self.embedding.embed_model.outputs)
        self.tf_model = keras.Model(self.embedding.embed_model.inputs, output)

    def compile_model(self,
                      optimizer: Any = None,
                      **kwargs: Any) -> None:
        if optimizer is None:
            optimizer = Adam(1e-5)

        self.tf_model.compile( optimizer=optimizer, **kwargs)

    def fit(self,
            x_train: TextSamplesVar,
            y_train: TextSamplesVar,
            x_validate: TextSamplesVar = None,
            y_validate: Union[ClassificationLabelVar, MultiLabelClassificationLabelVar] = None,
            *,
            batch_size: int = 64,
            epochs: int = 5,
            callbacks: List['keras.callbacks.Callback'] = None,
            fit_kwargs: Dict = None) -> 'keras.callbacks.History':
        train_gen = CorpusGenerator(x_train, y_train)
        if x_validate is not None:
            valid_gen = CorpusGenerator(x_validate,y_validate)
        else:
            valid_gen = None
        return self.fit_generator(train_sample_gen=train_gen,
                                  valid_sample_gen=valid_gen,
                                  batch_size=batch_size,
                                  epochs=epochs,
                                  callbacks=callbacks,
                                  fit_kwargs=fit_kwargs)

    def fit_generator(self,
                      train_sample_gen: CorpusGenerator,
                      valid_sample_gen: CorpusGenerator = None,
                      *,
                      batch_size: int = 64,
                      epochs: int = 5,
                      callbacks: List['keras.callbacks.Callback'] = None,
                      fit_kwargs: Dict = None) -> 'keras.callbacks.History':
        self.build_model_generator([g for g in [train_sample_gen, valid_sample_gen] if g])
        model_summary = []
        self.tf_model.summary(print_fn=lambda x: model_summary.append(x))
        logger.debug('\n'.join(model_summary))

        train_set = BertGeneratorDataSet(train_sample_gen,
                                 text_processor=self.text_processor,
                                 segment=self.embedding.segment,
                                 seq_length=self.sequence_length,
                                 batch_size=batch_size)

        if fit_kwargs is None:
            fit_kwargs = {}

        if valid_sample_gen:
            valid_gen = BertGeneratorDataSet(valid_sample_gen,
                                     text_processor=self.text_processor,
                                     segment=self.embedding.segment,
                                     seq_length=self.sequence_length,
                                     batch_size=batch_size)
            fit_kwargs['validation_data'] = valid_gen.take()
            fit_kwargs['validation_steps'] = len(valid_gen)

        if "steps_per_epoch" not in fit_kwargs:
            fit_kwargs["steps_per_epoch"] = len(train_set)
        return self.tf_model.fit(train_set.take(),
                                 #steps_per_epoch=len(train_set),
                                 epochs=epochs,
                                 callbacks=callbacks,
                                 **fit_kwargs)






import ahocorasick
import time


class AhocorasickNer:
    def __init__(self, user_dict_path):
        self.user_dict_path = user_dict_path
        self.actree = ahocorasick.Automaton()

    def add_keywords(self):
        flag = 0
        with open(self.user_dict_path, "r", encoding="utf-8") as file:
            for line in file:
                word, flag = line.strip(), flag + 1
                self.actree.add_word(word, (flag, word))
        self.actree.make_automaton()

    def get_ner_results(self, sentence):
        ner_results = []
        # i的形式为(index1,(index2,word))
        # index1: 提取后的结果在sentence中的末尾索引
        # index2: 提取后的结果在self.actree中的索引
        for i in self.actree.iter(sentence):
            ner_results.append((i[1], i[0] + 1 - len(i[1][1]), i[0] + 1))
        return ner_results


if __name__ == "__main__":
    test_paht = r'F:\nlp_data\word_dict\Organization-Names-Corpus.txt'
    ahocorasick_ner = AhocorasickNer(user_dict_path=test_paht)
    ahocorasick_ner.add_keywords()

    testList = [
        '我想去北京大学',
        "明天帮我去北京银行取100万",
        '四川省人民医院我去过'
    ]

    for sentence in testList:
        ss = time.time()
        res = ahocorasick_ner.get_ner_results(sentence)
        print("TIME  : {0}ms!".format(round(1000 * (time.time() - ss), 3)))
        print("OUTPUT:{0}".format(res))


# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements Focal loss."""

import tensorflow as tf
import tensorflow.keras.backend as K
from typeguard import typechecked

from tensorflow_addons.utils.keras_utils import  LossFunctionWrapper
from tensorflow_addons.utils.types import FloatTensorLike,TensorLike

@tf.keras.utils.register_keras_serializable(package='nlp_tools')
class SigmoidFocalCrossEntropy(LossFunctionWrapper):
    """Implements the focal loss function.
        Focal loss was first introduced in the RetinaNet paper
        (https://arxiv.org/pdf/1708.02002.pdf). Focal loss is extremely useful for
        classification when you have highly imbalanced classes. It down-weights
        well-classified examples and focuses on hard examples. The loss value is
        much high for a sample which is misclassified by the classifier as compared
        to the loss value corresponding to a well-classified example. One of the
        best use-cases of focal loss is its usage in object detection where the
        imbalance between the background class and other classes is extremely high.
        Usage:
        >>> fl = tfa.losses.SigmoidFocalCrossEntropy()
        >>> loss = fl(
        ...     y_true = [[1.0], [1.0], [0.0]],y_pred = [[0.97], [0.91], [0.03]])
        >>> loss
        <tf.Tensor: shape=(3,), dtype=float32, numpy=array([6.8532745e-06, 1.9097870e-04, 2.0559824e-05],
        dtype=float32)>
        Usage with `tf.keras` API:
        >>> model = tf.keras.Model()
        >>> model.compile('sgd', loss=tfa.losses.SigmoidFocalCrossEntropy())
        Args:
          alpha: balancing factor, default value is 0.25.
          gamma: modulating factor, default value is 2.0.
        Returns:
          Weighted loss float `Tensor`. If `reduction` is `NONE`, this has the same
              shape as `y_true`; otherwise, it is scalar.
        Raises:
            ValueError: If the shape of `sample_weight` is invalid or value of
              `gamma` is less than zero.
        """
    @typechecked
    def __init__(self,
                 from_logits: bool = False,
                 alpha:FloatTensorLike = 0.25,
                 gamma:FloatTensorLike = 2.0,
                 reduction: str = tf.keras.losses.Reduction.NONE,
                 name: str = 'sigmoid_focal_crossentropy'):
        super().__init__(
            sigmoid_focal_crossentropy,
            name=name,
            reduction=reduction,
            from_logits=from_logits,
            alpha=alpha,
            gamma=gamma
        )


def sigmoid_focal_crossentropy(
        y_true: TensorLike,
        y_pred: TensorLike,
        alpha: FloatTensorLike = 0.25,
        gamma: FloatTensorLike = 2.0,
        from_logits: bool = False
) -> tf.Tensor:
    """Implements the focal loss function.
        Focal loss was first introduced in the RetinaNet paper
        (https://arxiv.org/pdf/1708.02002.pdf). Focal loss is extremely useful for
        classification when you have highly imbalanced classes. It down-weights
        well-classified examples and focuses on hard examples. The loss value is
        much high for a sample which is misclassified by the classifier as compared
        to the loss value corresponding to a well-classified example. One of the
        best use-cases of focal loss is its usage in object detection where the
        imbalance between the background class and other classes is extremely high.
        Args:
            y_true: true targets tensor.
            y_pred: predictions tensor.
            alpha: balancing factor.
            gamma: modulating factor.
        Returns:
            Weighted loss float `Tensor`. If `reduction` is `NONE`,this has the
            same shape as `y_true`; otherwise, it is scalar.
        """
    if gamma and gamma < 0 :
        raise ValueError("Value of gamma should be greater than or equal to zero.")

    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true,dtype=y_pred.dtype)

    # Get the cross_entropy for each entry
    ce = K.binary_crossentropy(y_true,y_pred, from_logits=from_logits)

    # If logits are provided then convert the predictions into probabilities
    if from_logits:
        pred_prob = tf.sigmoid(y_pred)
    else:
        pred_prob = y_pred

    p_t = (y_true * pred_prob) + ((1 - y_true) * (1 - pred_prob))
    alpha_factor = 1.0
    modulating_factor = 1.0

    if alpha:
        alpha = tf.cast(alpha, dtype=y_true.dtype)
        alpha_factor = y_true * alpha + (1 - y_true) * (1 - alpha)

    if gamma:
        gamma = tf.cast(gamma, dtype=y_true.dtype)
        modulating_factor = tf.pow((1.0 - p_t),gamma)

    # compute the final loss and return
    return tf.reduce_sum(alpha_factor * modulating_factor * ce ,axis=-1)


import collections
import operator
from typing import List, Union, Dict, Optional, Any, Tuple

import numpy as np
import tqdm

from nlp_tools.generators import CorpusGenerator
from nlp_tools.processors.abc_label_processor import ABCLabelProcessor
from nlp_tools.types import  TextSamplesVar
from nlp_tools.utils import MultiLabelBinarizer

class ClassificationProcessor(ABCLabelProcessor):

    def to_dict(self) -> Dict[str, Any]:
        data = super(ClassificationProcessor, self).to_dict()
        data['config']['label2id'] = self.vocab2idx
        return data

    def __init__(self,
                 multi_label: bool = False,
                 label2id = None,
                 **kwargs: Any) -> None:

        super(ClassificationProcessor, self).__init__(**kwargs)
        self.vocab2idx = label2id
        self.multi_label = multi_label
        self.multi_label_binarizer = MultiLabelBinarizer(self.vocab2idx)

    def build_vocab_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        if not self.vocab2idx:
            vocab2idx: Dict[str, int] = {}
            token2count: Dict[str, int] = {}
            for generator in generators:
                if self.multi_label:
                    for (_, label) in tqdm.tqdm(generator.sample(), desc="Preparing classification label vocab dict"):
                        for token in label:
                            count = token2count.get(token, 0)
                            token2count[token] = count + 1
                else:
                    for _, label in tqdm.tqdm(generator.sample(), desc="Preparing classification label vocab dict"):
                        count = token2count.get(label, 0)
                        token2count[label] = count + 1

            sorted_token2count = sorted(token2count.items(),
                                        key=operator.itemgetter(0))
            token2count = collections.OrderedDict(sorted_token2count)

            for token, token_count in token2count.items():
                if token not in vocab2idx:
                    vocab2idx[token] = len(vocab2idx)
            self.vocab2idx = vocab2idx
        self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])
        self.multi_label_binarizer = MultiLabelBinarizer(self.vocab2idx)

    def get_tensor_shape(self, batch_size: int, seq_length: int) -> Tuple:
        if self.multi_label:
            return batch_size, len(self.vocab2idx)
        else:
            return (batch_size,)

    def transform(self,
                  samples: TextSamplesVar,
                  *,
                  seq_length: int = None,
                  max_position: int = None,
                  segment: bool = False) -> np.ndarray:
        if self.multi_label:
            sample_tensor = self.multi_label_binarizer.transform(samples)
            return sample_tensor

        sample_tensor = [self.vocab2idx[i] for i in samples]
        return np.array(sample_tensor)

    def inverse_transform(self,  # type: ignore[override]
                          labels: Union[List[int], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> Union[List[List[str]], List[str]]:
        if self.multi_label:
            return self.multi_label_binarizer.inverse_transform(labels,
                                                                threshold=threshold)
        else:
            return [self.idx2vocab[i] for i in labels]


if __name__ == "__main__":
    pass



from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer


class DoNothingTokenizer(ABCTokenizer):
    """list tokenizer
    """

    def __init__(self):
        pass

    def tokenize(self, text: str, **kwargs):
        return text

    def encode(self, text: str, **kwargs):
        return text




# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Discriminative Layer Training Optimizer for TensorFlow."""

from typing import List,Union

import tensorflow as tf
from typeguard import typechecked




@tf.keras.utils.register_keras_serializable(package='nlp_tools')
class MultiOptimizer(tf.keras.optimizers.Optimizer):
    """Multi Optimizer Wrapper for Discriminative Layer Training.

    Creates a wrapper around a set of instantiated optimizer layer pairs.
    Generally useful for transfer learning of deep networks.

    Each optimizer will optimize only the weights associated with its paired layer.
    This can be used to implement discriminative layer training by assigning
    different learning rates to each optimizer layer pair.
    `(tf.keras.optimizers.Optimizer, List[tf.keras.layers.Layer])` pairs are also supported.
    Please note that the layers must be instantiated before instantiating the optimizer.

    Args:
        optimizers_and_layers: a list of tuples of an optimizer and a layer or model.
            Each tuple should contain exactly 1 instantiated optimizer and 1 object that
            subclasses `tf.keras.Model`, `tf.keras.Sequential` or `tf.keras.layers.Layer`.
            Nested layers and models will be automatically discovered.
            Alternatively, in place of a single layer, you can pass a list of layers.
        optimizer_specs: specialized list for serialization.
            Should be left as None for almost all cases.
            If you are loading a serialized version of this optimizer,
            please use `tf.keras.models.load_model` after saving a model compiled with this optimizer.

    Usage:

    >>> model = tf.keras.Sequential([
    ...     tf.keras.Input(shape=(4,)),
    ...     tf.keras.layers.Dense(8),
    ...     tf.keras.layers.Dense(16),
    ...     tf.keras.layers.Dense(32),
    ... ])
    >>> optimizers = [
    ...     tf.keras.optimizers.Adam(learning_rate=1e-4),
    ...     tf.keras.optimizers.Adam(learning_rate=1e-2)
    ... ]
    >>> optimizers_and_layers = [(optimizers[0], model.layers[0]), (optimizers[1], model.layers[1:])]
    >>> optimizer = tfa.optimizers.MultiOptimizer(optimizers_and_layers)
    >>> model.compile(optimizer=optimizer, loss="mse")

    Reference:
        - [Universal Language Model Fine-tuning for Text Classification](https://arxiv.org/abs/1801.06146)
        - [Collaborative Layer-wise Discriminative Learning in Deep Neural Networks](https://arxiv.org/abs/1607.05440)

    Note: Currently, `tfa.optimizers.MultiOptimizer` does not support callbacks that modify optimizers.
        However, you can instantiate optimizer layer pairs with
        `tf.keras.optimizers.schedules.LearningRateSchedule`
        instead of a static learning rate.

    This code should function on CPU, GPU, and TPU. Apply with `tf.distribute.Strategy().scope()` context as you
    would with any other optimizer.
    """
    @typechecked
    def __init__(self,
                 optimizers_and_layers: Union[list,None] = None,
                 optimizer_specs: Union[list,None] = None,
                 name: str = 'MultiOptimizer',
                 **kwargs):
        super(MultiOptimizer,self).__init__(name,**kwargs)

        if optimizer_specs is None and optimizers_and_layers is not None:
            self.optimizer_specs = [
                self.create_optimizer_spec(optimizer,layers_or_model) for
                optimizer ,layers_or_model in optimizers_and_layers
            ]
        elif optimizer_specs is not None and optimizers_and_layers is None:
            self.optimizer_specs = [
                self.maybe_initialize_optimizer_spec(spec) for spec in optimizer_specs
            ]

        else:
            raise RuntimeError(
                'Must specify one of `optimizers_and_layers` or `optimizer_specs`.'
            )

    def apply_gradients(self,grads_and_vars,**kwargs):
        """Wrapped apply_gradient method.

                Returns an operation to be executed.
                """
        for spec in self.optimizer_specs:
            spec['gv'] = []

        for grad,var in tuple(grads_and_vars):
            for spec in self.optimizer_specs:
                for name in spec['weights']:
                    if var.name == name:
                        spec['gv'].append((grad,var))

        return tf.group([
            spec["optimizer"].apply_gradients(spec['gv'],**kwargs)
            for spec in self.optimizer_specs
        ])

    def get_config(self):
        config = super(MultiOptimizer,self).get_config()
        for optimizer_spec in self.optimizer_specs:
            if 'gv' in optimizer_spec.keys():
                del optimizer_spec['gv']
        config.update({"optimizer_specs": self.optimizer_specs})
        return config




    @classmethod
    def create_optimizer_spec(cls,
                              optimizer: tf.keras.optimizers.Optimizer,
                              layers_or_model: Union[
                                  tf.keras.Model,
                                  tf.keras.Sequential,
                                  tf.keras.layers.Layer,
                                  List[tf.keras.layers.Layer]
                              ]):
        """Creates a serializable optimizer spec.

                The name of each variable is used rather than `var.ref()` to enable serialization and deserialization.
                """
        if isinstance(layers_or_model, list):
            weights = [
                var.name for sublayer in layers_or_model for var in sublayer.weights
            ]
        else:
            weights = [var.name for var in layers_or_model.weights]

        return {
            "optimizer":optimizer,
            'weights':weights
        }

    @classmethod
    def maybe_initialize_optimizer_spec(cls,optimizer_spec):
        if isinstance(optimizer_spec['optimizer'],dict):
            optimizer_spec['optimizer'] = tf.keras.optimizers.deserialize(optimizer_spec['optimizer'])

        return optimizer_spec

    def __repr__(self):
        return "multi optimizer with %i optimizer layer pairs" % len(self.optimizer_specs)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   f1score_save_callback.py
@Contact :   544855237@qq.com
@License :   (C)Copyright 2017-2018, Liugroup-NLPR-CASIA

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/12/20 下午3:07   qiufengfeng      1.0         None
'''
import json
import os.path

from tensorflow.keras.callbacks import Callback
from sklearn.metrics import f1_score,classification_report
from nlp_tools.metrics.plot.confusion_seaborn import plot_confusion_mat

class F1SaveCallback(Callback):
    def __init__(self,model_check,model_save_path,valid_data,batch_size=32,label_names=None,early_stop_num=15 ):
        """

        @param model_check:
        @param model_save_path:
        @param valid_data:
        @param batch_size:
        @param label_names:
        @param early_stop_num:
        """
        self.model_check = model_check
        self.model_save_path = model_save_path
        self.batch_size= batch_size
        self.label_names = label_names
        self.early_stop_num = early_stop_num
        self.valid_x = [x[0] for x in valid_data]
        self.valid_y_ids = model_check.label_processor.transform([x[1] for x in valid_data])


        self.best_val_f1 = 0
        self.count = 0


        self.confusion_save_path = os.path.join(self.model_save_path,'confusion.jpg')
        self.classify_report_path = os.path.join(self.model_save_path,'classify_report.txt')



    def caculate_f1(self):
        predict_labels  = self.model_check.predict(self.valid_x,batch_size=self.batch_size,return_labels=False)
        f1_score_value = f1_score(self.valid_y_ids,predict_labels,average='macro')
        report = classification_report(self.valid_y_ids,predict_labels,target_names=self.label_names,)
        print(report)
        figure = plot_confusion_mat(self.valid_y_ids, predict_labels, self.label_names)
        if f1_score_value >= self.best_val_f1:
            self.best_val_f1 = f1_score_value
            self.model_check.save(self.model_save_path)
            self.count = 0

            figure.savefig(self.confusion_save_path, dpi=600)
            with open(self.classify_report_path,'w',encoding='utf-8') as fwrite:
                if type(report) == dict:
                    report = json.dumps(report,indent=4,ensure_ascii=False)
                fwrite.write(report)

            print("new best f1: %.5f\n" % f1_score_value )
        else:
            print(
                'valid:  f1: %.5f, best f1: %.5f\n' %
                (f1_score_value, self.best_val_f1)
            )







    def on_epoch_end(self, epoch, logs=None):
        self.caculate_f1()
        self.count += 1
        if self.count >self.early_stop_num:
            self.model.stop_training = True
            self.count = 0


import os
from typing import Dict,Any

from nlp_tools.embeddings.transformer_embedding import TransformerEmbedding

class BertEmbedding(TransformerEmbedding):
    """
    BertEmbedding is a simple wrapped class of TransformerEmbedding.
    If you need load other kind of transformer based language model, please use the TransformerEmbedding.
    """

    def to_dict(self) -> Dict[str, Any]:
        info_dic = super(BertEmbedding, self).to_dict()
        return info_dic

    def __init__(self,
                 model_folder: str = None,
                 model_type='bert',
                 **kwargs: Any):
        """

        Args:
            model_folder: path of checkpoint folder.
            kwargs: additional params
        """
        if model_folder:
            if model_type == 'electra':
                config_path = os.path.join(model_folder, 'bert_config.json')
                checkpoint_path = os.path.join(model_folder, 'electra_base')
            else:
                config_path = os.path.join(model_folder, 'bert_config.json')
                checkpoint_path = os.path.join(model_folder, 'bert_model.ckpt')
                if not os.path.exists(checkpoint_path):
                    checkpoint_path = None
            kwargs['checkpoint_path'] = checkpoint_path
            kwargs['config_path'] = config_path
        kwargs['model_type'] = model_type
        super(BertEmbedding, self).__init__(**kwargs)






import tensorflow.keras.backend as K
import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package='nlp_tools')
def multilabel_categorical_crossentropy(y_true,y_pred):
    """多标签分类的交叉熵
        说明：
            1. y_true和y_pred的shape一致，y_true的元素非0即1，
               1表示对应的类为目标类，0表示对应的类为非目标类；
            2. 请保证y_pred的值域是全体实数，换言之一般情况下
               y_pred不用加激活函数，尤其是不能加sigmoid或者
               softmax；
            3. 预测阶段则输出y_pred大于0的类；
            4. 详情请看：https://kexue.fm/archives/7359 。
        """

    y_pred = (1 -2 *y_true) * y_pred
    y_pred_neg = y_pred - y_true * 1e12
    y_pred_pos = y_pred - (1 - y_true) * 1e12
    zeros = K.zeros_like(y_pred[...,:1])
    y_pred_neg = K.concatenate([y_pred_neg,zeros],axis=-1)
    y_pred_pos = K.concatenate([y_pred_pos,zeros],axis=-1)
    neg_loss = tf.reduce_logsumexp(y_pred_neg,axis=-1)
    pos_loss = tf.reduce_logsumexp(y_pred_pos,axis=-1)
    return neg_loss + pos_loss

@tf.keras.utils.register_keras_serializable(package='nlp_tools')
def global_pointer_crossentropy(y_true,y_pred):
    """给GlobalPointer设计的交叉熵
        其实就是将多维度的多分类【B,N，R,C】 变成 【B*N*R,C】，然后在调用multilabel_categorical_crossentropy
        """
    bh = K.prod(K.shape(y_pred)[:2])
    y_true = K.reshape(y_true,(bh, -1))
    y_pred = K.reshape(y_pred,(bh, -1))
    return K.mean(multilabel_categorical_crossentropy(y_true,y_pred))




from tensorflow.keras.losses import kullback_leibler_divergence as kld
from tensorflow.keras import backend as K
import tensorflow as tf


class RDropLoss():
    # https://spaces.ac.cn/archives/8496
    __name__ ="r_drop_loss"
    def __init__(self,loss_func=None):
        self.loss_func = loss_func

    def __call__(self,y_true, y_pred,alpha=4):
        """配合上述生成器的R-Drop Loss
            其实loss_kl的除以4，是为了在数量上对齐公式描述结果。
            """
        loss = self.loss_func(y_true, y_pred)  # 原来的loss
        loss_kl = kld(y_pred[::2], y_pred[1::2]) + kld(y_pred[1::2], y_pred[::2])
        return loss + K.mean(loss_kl) / 4 * alpha




from typing import Dict, Any

from tensorflow import keras
from nlp_tools.layers import L, KConditionalRandomField
from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel




class BiLSTM_CRF_Model(ABCLabelingModel):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_blstm': {
                'units': 64,
                'return_sequences': True
            },
            'layer_dropout': {
                'rate': 0.5
            },
            'layer_time_distributed': {},
        }



    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        crf = KConditionalRandomField()
        layer_stack = [
            L.Bidirectional(L.LSTM(**config['layer_blstm']), name='layer_blstm'),
            L.Dropout(**config['layer_dropout'], name='layer_dropout'),
            L.Dense(output_dim, **config['layer_time_distributed']),
            crf
        ]

        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model = keras.Model(embed_model.inputs, tensor)
        self.crf_layer = crf

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            loss = self.crf_layer.loss
        if metrics is None:
            metrics = [self.crf_layer.accuracy]
        super(BiLSTM_CRF_Model, self).compile_model(loss=loss,
                                                    optimizer=optimizer,
                                                    metrics=metrics,
                                                    **kwargs)





from tensorflow.keras.utils import get_file
from typing import Tuple, List
from nlp_tools import macros as k
import os
from nlp_tools.corpus import DataReader
from nlp_tools import utils
import logging
import pandas as pd

class SMP2018ECDTCorpus(object):
    """
    https://worksheets.codalab.org/worksheets/0x27203f932f8341b79841d50ce0fd684f/

    This dataset is released by the Evaluation of Chinese Human-Computer Dialogue Technology (SMP2018-ECDT)
    tasks 1 and is provided by the iFLYTEK Corporation, which is a Chinese human-computer dialogue dataset.
    sample::

              label           query
        0   weather        今天东莞天气如何
        1       map  从观音桥到重庆市图书馆怎么走
        2  cookbook          鸭蛋怎么腌？
        3    health         怎么治疗牛皮癣
        4      chat             唠什么
    """

    __corpus_name__ = 'SMP2018ECDTCorpus'
    __zip_file__name = 'http://s3.bmio.net/kashgari/SMP2018ECDTCorpus.tar.gz'

    @classmethod
    def load_data(cls,
                  subset_name: str = 'train',
                  shuffle: bool = True,
                  cutter: str = 'char') -> Tuple[List[List[str]], List[str]]:
        """
        Load dataset as sequence classification format, char level tokenized

        features: ``[['听', '新', '闻', '。'], ['电', '视', '台', '在', '播', '什', '么'], ...]``

        labels: ``['news', 'epg', ...]``

        Samples::
            train_x, train_y = SMP2018ECDTCorpus.load_data('train')
            test_x, test_y = SMP2018ECDTCorpus.load_data('test')

        Args:
            subset_name: {train, test, valid}
            shuffle: should shuffle or not, default True.
            cutter: sentence cutter, {char, jieba}

        Returns:
            dataset_features and dataset labels
        """

        corpus_path = get_file(cls.__corpus_name__,
                               cls.__zip_file__name,
                               cache_dir=k.DATA_PATH,
                               untar=True)

        if cutter not in ['char', 'jieba', 'none']:
            raise ValueError('cutter error, please use one onf the {char, jieba}')

        df_path = os.path.join(corpus_path, f'{subset_name}.csv')
        df = pd.read_csv(df_path)
        if cutter == 'jieba':
            try:
                import jieba
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "please install jieba, `$ pip install jieba`")
            x_data = [list(jieba.cut(item)) for item in df['query'].to_list()]
        elif cutter == 'char':
            x_data = [list(item) for item in df['query'].to_list()]
        else:
            x_data = [item for item in df['query'].to_list()]
        y_data = df['label'].to_list()

        if shuffle:
            x_data, y_data = utils.unison_shuffled_copies(x_data, y_data)
        logging.debug(f"loaded {len(x_data)} samples from {df_path}. Sample:\n"
                      f"x[0]: {x_data[0]}\n"
                      f"y[0]: {y_data[0]}")
        return x_data, y_data


if __name__ == "__main__":
    a, b = SMP2018ECDTCorpus.load_data(cutter='none')
    print(a[:2])
    print(b[:2])
    print("Hello world")


from transformers import AutoConfig

class PrefixPromptConfig(AutoConfig):
    def __init__(self):
        super().__init__()
        self.prefix_hidden_size = None
        self.pre_seq_len = None
        self.prefix_projection = None

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path,pre_seq_len=8,prefix_hidden_size=512,prefix_projection=True, **kwargs):
        config_obj = AutoConfig.from_pretrained(pretrained_model_name_or_path)
        config_obj.pre_seq_len = pre_seq_len
        config_obj.prefix_hidden_size = prefix_hidden_size
        config_obj.prefix_projection = prefix_projection
        return config_obj



if __name__ == '__main__':
    model_name = "roberta-large"

    config = PrefixPromptConfig.from_pretrained(model_name)
    print(config)


from typing import Tuple, List
import glob
import os
from tqdm import tqdm
from nlp_tools.generators import ABCGenerator

class AutoTitleThunc(ABCGenerator):
    __corpus_name__ = 'thunc_text_summarization'

    def __init__(self,file_path,shuffle: bool = True,buffer_size: int = 5000):
        self.file_path = file_path
        self.shufflle = shuffle
        super(AutoTitleThunc, self).__init__(buffer_size=buffer_size)

    def __iter__(self):

        golb_file_path = os.path.join(self.file_path,"*/*.txt")
        txts = glob.glob(golb_file_path)


        for txt in tqdm(txts):
            text = open(txt, encoding='utf-8').read()
            text = text.split('\n')
            if len(text) > 1:
                title = text[0]
                content = '\n'.join(text[1:])
                yield content,title
                #titles.append(title)
                #contexts.append(content)
       # return contexts,titles

    def __len__(self) -> int:
        return 890000


    def load_data(self):

        golb_file_path = os.path.join(self.file_path,"*/*.txt")
        txts = glob.glob(golb_file_path)

        titles = []
        contexts = []
        count = 0
        for txt in tqdm(txts):
            text = open(txt, encoding='utf-8').read()
            text = text.split('\n')
            if len(text) > 1:
                title = text[0]
                content = '\n'.join(text[1:])
                titles.append(title)
                contexts.append(content)

            count += 1

        return contexts,titles


# encoding: utf-8


import collections
import operator
from typing import Dict, List, Any, Union

import numpy as np
import tqdm
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.processors.abc_label_processor import ABCLabelProcessor
from nlp_tools.utils.ner_utils import get_entities
from tensorflow.keras.utils import to_categorical


class CascadeSentenceLabelProcessor(ABCLabelProcessor):
    """
    Generic processors for the sequence samples.
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(CascadeSentenceLabelProcessor, self).to_dict()
        data['config'].update({
            'embedding_max_position': self.embedding_max_position,
            'max_sentence_length':self.max_sentence_length,
            "segment_mode":self.segment_mode
        })
        return data

    def __init__(self,
                 segment_mode='BIES',
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        super(CascadeSentenceLabelProcessor, self).__init__(**kwargs)
        self._initial_vocab_dic = {
            "O":0
        }
        self.segment_mode = segment_mode
        if self.segment_mode == "BIES":
            self.segment_labels = {
                "O":0,
                "B":1,
                "I":2,
                "E":3,
                "S":4,
            }
        else:
            self.segment_labels = {
                'O':0,
                "B":1,
                "I":2,
            }
        self.segmetn_idx_2_label = {v:k for k,v in self.segment_labels.items()}


    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length

    def build_vocab_generator(self,generators) -> None:
        if not self.vocab2idx:
            tag_set = set()
            vocab2idx = self._initial_vocab_dic
            for gen in generators:
                for (_, label_list) in tqdm.tqdm(gen.sample(), desc="Preparing label dict"):
                    ## ner任务下面，label 应该是一个list
                    assert type(label_list) == list
                    for sublabel in label_list:
                        label = sublabel[2]
                        if label not in tag_set:
                            tag_set.add(label)


            for label in list(set(tag_set)):
                vocab2idx[label] = len(vocab2idx)
            self.vocab2idx = vocab2idx
            self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])




    def transform(self,samples,seq_length=None) -> np.ndarray:
        if not seq_length:
            seq_length = self.max_sentence_length

        if not seq_length:
            seq_length = self.embedding_max_position

        if self.embedding_max_position is not None and self.embedding_max_position < seq_length:
            seq_length = self.embedding_max_position

        labels_numpy = np.zeros((len(samples), seq_length))
        segment_numpy = np.zeros((len(samples), seq_length))

        for index, sample_labels in enumerate(samples):
            for sub_label in sample_labels:
                (start_position, end_position, label_type) = sub_label
                if start_position >= seq_length:
                    continue

                if end_position >= seq_length:
                    continue

                labels_numpy[index, start_position:end_position+1] = self.vocab2idx[ label_type]
                if self.segment_mode == 'BI':
                    segment_numpy[index,start_position] = self.segment_labels['B']
                    for i in range(start_position+1,end_position+1):
                        segment_numpy[index,i] = self.segment_labels['I']
                else:
                    if end_position - start_position == 0:
                        segment_numpy[index, start_position] = self.segment_labels['S']
                    else:
                        segment_numpy[index, start_position] = self.segment_labels['B']
                        for i in range(start_position + 1, end_position + 1):
                            segment_numpy[index,i] = self.segment_labels['I']
                        segment_numpy[index,end_position] = self.segment_labels['E']

        return (segment_numpy,labels_numpy)

    def inverse_transform(self,
                          labels_and_segment,
                          lengths: List[int] = None,
                          mapping_list = None,**kwargs ):
        preds,pred_labels = labels_and_segment

        preds = [[self.segmetn_idx_2_label[i] for i in p] for p in preds]
        pred_labels = [ [self.idx2vocab[i] for i in p] for p in pred_labels]

        # temp_labels =[]
        # for p,q in zip(preds,pred_labels):
        #     temp_labels.append(self.trans_label(p,q))
        temp_labels = self.trans_label(preds,pred_labels)

        result = []
        for index, (label, sequece_length) in enumerate(zip(temp_labels, lengths)):
            label = label[:sequece_length]
            if len(np.shape(label)) == 2:
                label = np.argmax(label,axis=-1)
            #label = [self.idx2vocab[i] for i in label]

            entities = get_entities(label)

            # 将position映射为真实的句子中的index，因为之前由于分词以后，会影响index
            format_entities = []
            for (tagname,start,end) in entities:
                if mapping_list:
                    if mapping_list[index][start] and mapping_list[index][end]:
                        format_entities.append((mapping_list[index][start][0], mapping_list[index][end][-1], tagname))
                else:
                    format_entities.append((start + 1, end + 1, tagname))

            result.append(format_entities)
        return result

    def trans_label(self,segment_labels_id, attr_labels_id):
        """
        答案拼接2种方式：
        例子：
        bmeo：[B, M, E]
        attr: [LOC, LOC, ORG]
        1. 实体词 bmeo 与 attr 的属性全都对应起来
        result: [B-LOC, M-LOC, E-ORG]
        2. 实体词 bmeo 的e所对应的attr作为该实体的attr
        result: [B-ORG, M-ORG, E-ORG]
        目前采用第一种
        """
        std_labels = []
        for index, bmeo_line in enumerate(segment_labels_id):
            bmeo_attr_label = []
            attr_line = attr_labels_id[index]
            for item in list(zip(bmeo_line, attr_line)):
                bmeo_id = item[0]
                attr_id = item[1]
                if bmeo_id == "O":
                    bmeo_attr = "O"
                else:
                    bmeo_attr = bmeo_id + "-" + attr_id
                bmeo_attr_label.append(bmeo_attr)
            std_labels.append(bmeo_attr_label)
        # print("std labels")
        # print(std_labels)
        return std_labels





if __name__ == "__main__":
    pass



import tensorflow as tf

from nlp_tools.embeddings.abc_embedding import ABCEmbedding
from nlp_tools.layers import L

class AttGRUDecoder(tf.keras.Model):
    def __init__(self,
                 embedding: ABCEmbedding,
                 vocab_size: int,
                 hidden_size: int = 1024):
        super(AttGRUDecoder, self).__init__()
        self.embedding = embedding
        self.hidden_size = hidden_size
        self.gru = tf.keras.layers.GRU(hidden_size,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')
        self.fc = tf.keras.layers.Dense(vocab_size)

        # 用于注意力
        self.attention = L.BahdanauAttention(hidden_size)

    def call(self, x, hidden, enc_output):
        # enc_output shape == (batch_size, max_length, hidden_size)
        context_vector, attention_weights = self.attention(hidden, enc_output)

        if self.embedding.segment:
            x = x, tf.zeros(x.shape)

        # x shape after passing through embedding == (batch_size, 1, embedding_dim)
        x = self.embedding.embed_model(x)

        # x shape after concatenation == (batch_size, 1, embedding_dim + hidden_size)
        x = tf.concat([tf.expand_dims(context_vector, 1), x], axis=-1)

        # passing the concatenated vector to the GRU
        output, state = self.gru(x)

        # output shape == (batch_size * 1, hidden_size)
        output = tf.reshape(output, (-1, output.shape[2]))

        # output shape == (batch_size, vocab)
        x = self.fc(output)

        return x, state, attention_weights

    def model(self):
        x1 = L.Input(shape=(None,))
        x2 = L.Input(shape=(self.hidden_size,))
        x3 = L.Input(shape=(self.hidden_size,))
        return tf.keras.Model(inputs=[x1, x2, x3],
                              outputs=self.call(x1, x2, x3),
                              name='AttGRUDecoder')


if __name__ == "__main__":
    pass



# encoding: utf-8


import collections
import operator
from typing import Dict, List, Any, Union

import numpy as np
import tqdm
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.processors.abc_label_processor import ABCLabelProcessor
from nlp_tools.types import TextSamplesVar


class GlobalPointerLabelProcessor(ABCLabelProcessor):
    """
    Generic processors for the sequence samples.
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(GlobalPointerLabelProcessor, self).to_dict()
        data['config'].update({
            'embedding_max_position': self.embedding_max_position,
            'max_sentence_length':self.max_sentence_length
        })
        return data

    def __init__(self,
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        super(GlobalPointerLabelProcessor, self).__init__(**kwargs)
        self._initial_vocab_dic = {
        }



    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length

    def build_vocab_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        if not self.vocab2idx:
            vocab2idx = self._initial_vocab_dic
            token2count: Dict[str, int] = {}
            for gen in generators:
                for (_, label_list) in tqdm.tqdm(gen.sample(), desc="Preparing label dict"):
                    ## ner任务下面，label 应该是一个list
                    assert type(label_list) == list
                    for sublabel in label_list:
                        label = sublabel[2]
                        count = token2count.get(sublabel[2], 0)
                        token2count[label] = count + 1

            sorted_token2count = sorted(token2count.items(),
                                        key=operator.itemgetter(1),
                                        reverse=True)
            token2count = collections.OrderedDict(sorted_token2count)
            for token, token_count in token2count.items():
                vocab2idx[token] = len(vocab2idx)
            self.vocab2idx = vocab2idx
            self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])





    def transform(self,samples,seq_length=None) -> np.ndarray:
        if not seq_length:
            seq_length = self.max_sentence_length

        if not seq_length:
            seq_length = self.embedding_max_position

        if self.embedding_max_position is not None and self.embedding_max_position < seq_length:
            seq_length = self.embedding_max_position

        labels_numpy = np.zeros((len(samples),len(self.vocab2idx), seq_length, seq_length))
        for index, sample_labels in enumerate(samples):
            for sub_label in sample_labels:
                (start_position, end_position, label_type) = sub_label
                label2index = self.vocab2idx[label_type]

                if start_position >= seq_length:
                    continue

                if end_position >= seq_length:
                    continue

                labels_numpy[index,label2index,start_position,end_position] = 1
        return labels_numpy

    def inverse_transform(self, labels, lengths=None, mapping_list=None, threshold=0, return_max_entity=False):
        result = []
        if return_max_entity:
            for index,(scores ,sequece_length )in enumerate(zip(labels,lengths)):
                entities = []
                scores = scores[:,:sequece_length,:sequece_length]
                scores[:, [0, -1]] -= np.inf  # 去除添加的[cls]和【seq】这一行的判断
                scores[ :, :, [0, -1]] -= np.inf  # 去除每一行的首尾添加的[cls]和【seq】
                scores = np.transpose(scores,[1,2,0])
                max_scores = np.max(scores,axis=-1)
                for (start,end,l) in zip(*np.where(scores > threshold)):
                    if max_scores[start, end] == scores[start, end, l]:
                        if mapping_list:
                            entities.append(
                                (mapping_list[index][start][0], mapping_list[index][end][-1], self.idx2vocab[l])
                            )
                        else:
                            # 如果mapping为空，start,end都要减一，因为在tokenizer的时候增加了[cls]，所以这边位置要减1
                            entities.append(
                                (start - 1, end - 1, self.idx2vocab[l])
                            )
                result.append(entities)

        else:
            for index,(scores ,sequece_length )in enumerate(zip(labels,lengths)):
                entities = []
                scores = scores[:,:sequece_length,:sequece_length]
                scores[:, [0, -1]] -= np.inf  # 去除添加的[cls]和【seq】这一行的判断
                scores[ :, :, [0, -1]] -= np.inf  # 去除每一行的首尾添加的[cls]和【seq】
                for l, start, end in zip(*np.where(scores > threshold)):
                    if mapping_list:
                        entities.append(
                            (mapping_list[index][start][0], mapping_list[index][end][-1], self.idx2vocab[l])
                        )
                    else:
                        # 如果mapping为空，start,end都要减一，因为在tokenizer的时候增加了[cls]，所以这边位置要减1
                        entities.append(
                            (start - 1, end - 1, self.idx2vocab[l])
                        )
                result.append(entities)
        return result





if __name__ == "__main__":
    import numpy as np
    a = np.arange(48).reshape((3,4,4))
    a = a.astype(np.float64)
    print(a)
    b = np.where(a > 3)
    print(b)

    for l, start, end in zip(*np.where(a > 3)):
        print(l,start,end)



from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer


class JiebaTokenizer(ABCTokenizer):
    """Jieba tokenizer
    """

    def __init__(self):
        try:
            import jieba
            self._jieba = jieba
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Jieba module not found, please install use `pip install jieba`")

    def tokenize(self, text: str, **kwargs):
        """
        Tokenize text into token sequence
        Args:
            text: target text sample

        Returns:
            List of tokens in this sample
        """

        return list(self._jieba.cut(text, **kwargs))

    def encode(self, text: str, **kwargs):
        """
        Tokenize text into token sequence
        Args:
            text: target text sample

        Returns:
            List of tokens in this sample
        """

        return list(self._jieba.cut(text, **kwargs))



# encoding: utf-8


import collections
import operator
from typing import Dict, List, Any, Union

import numpy as np
import tqdm
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.processors.abc_label_processor import ABCLabelProcessor
from nlp_tools.utils.ner_utils import get_entities
from tensorflow.keras.utils import to_categorical


class NerSentenceLabelProcessor(ABCLabelProcessor):
    """
    Generic processors for the sequence samples.
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(NerSentenceLabelProcessor, self).to_dict()
        data['config'].update({
            'embedding_max_position': self.embedding_max_position,
            'max_sentence_length':self.max_sentence_length,
            "tag_mode":self.tag_mode,
            "return_one_hot":self.return_one_hot
        })
        return data

    def __init__(self,
                 tag_mode='BI',
                 return_onehot = False,
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        super(NerSentenceLabelProcessor, self).__init__(**kwargs)
        self._initial_vocab_dic = {
            "O": 0
        }
        self.tag_mode = tag_mode
        self.return_one_hot = return_onehot

    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length

    def build_vocab_generator(self,generators) -> None:
        if not self.vocab2idx:
            tag_set = set()
            vocab2idx = self._initial_vocab_dic
            for gen in generators:
                for (_, label_list) in tqdm.tqdm(gen.sample(), desc="Preparing label dict"):
                    ## ner任务下面，label 应该是一个list
                    assert type(label_list) == list
                    for sublabel in label_list:
                        label = sublabel[2]
                        if label not in tag_set:
                            tag_set.add(label)


            for label in list(set(tag_set)):
                if self.tag_mode == "BI":
                    vocab2idx["B-" + label] = len(vocab2idx)
                    vocab2idx["I-" + label] = len(vocab2idx)
                else:
                    raise ValueError("不支持的tag_mode")
            self.vocab2idx = vocab2idx
            self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])




    def transform(self,samples,seq_length=None):
        if not seq_length:
            seq_length = self.max_sentence_length

        if not seq_length:
            seq_length = self.embedding_max_position

        if self.embedding_max_position is not None and self.embedding_max_position < seq_length:
            seq_length = self.embedding_max_position

        labels_numpy = np.zeros((len(samples), seq_length))

        for index, sample_labels in enumerate(samples):
            for sub_label in sample_labels:
                (start_position, end_position, label_type) = sub_label
                if start_position >= seq_length:
                    continue

                if end_position >= seq_length:
                    continue

                labels_numpy[index, start_position] = self.vocab2idx["B-" + label_type]
                if self.tag_mode == "BI":
                    for sub_index in range(start_position +1,end_position+1):
                        labels_numpy[index, sub_index] = self.vocab2idx["I-" + label_type]
        if self.return_one_hot:
            labels_numpy = to_categorical(labels_numpy,num_classes=self.vocab_size)
        return labels_numpy

    def inverse_transform(self,
                          labels: Union[List[List[int]], np.ndarray],
                          lengths: List[int] = None,
                          mapping_list = None ):
        result = []
        for index, (label, sequece_length) in enumerate(zip(labels, lengths)):
            label = label[:sequece_length]
            if len(np.shape(label)) == 2:
                label = np.argmax(label,axis=-1)
            label = [self.idx2vocab[i] for i in label]

            entities = get_entities(label)

            # 将position映射为真实的句子中的index，因为之前由于分词以后，会影响index
            format_entities = []
            for (tagname,start,end) in entities:
                if mapping_list:
                    if mapping_list[index][start] and mapping_list[index][end]:
                        format_entities.append((mapping_list[index][start][0], mapping_list[index][end][-1], tagname))
                else:
                    format_entities.append((start + 1, end + 1, tagname))

            result.append(format_entities)
        return result





if __name__ == "__main__":
    pass



import os
import logging

logger = logging.Logger('nlp_tools',level='DEBUG')
stream_handler = logging.StreamHandler()

if os.environ.get("NLPTOOLS_DEV") == 'True' :
    log_format =  '%(asctime)s [%(levelname)s] %(name)s:%(filename)s:%(lineno)d - %(message)s'
else:
    log_format = '%(asctime)s [%(levelname)s] %(name)s - %(message)s'

stream_handler.setFormatter(logging.Formatter(log_format))
logger.addHandler(stream_handler)



from typing import Dict, Any

from tensorflow import keras
import tensorflow as tf
import numpy as np

from nlp_tools.layers import L, KConditionalRandomField
from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel
from nlp_tools.logger import logger
from nlp_tools.layers.pool_layer import PoolerInputLayer

from tensorflow.keras.metrics import sparse_categorical_accuracy





class Cascade_Model(ABCLabelingModel):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_blstm': {
                'units': 64,
                'return_sequences': True
            },
            'layer_dropout': {
                'rate': 0.5
            },
            'layer_time_distributed': {},
        }



    def build_model_arc(self) -> None:

        segment_output_dim = len(self.label_processor.segmetn_idx_2_label.items())

        label_output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        crf = KConditionalRandomField()
        layer_stack = [
            L.Bidirectional(L.LSTM(**config['layer_blstm']), name='layer_blstm'),
            L.Dropout(**config['layer_dropout'], name='layer_dropout'),
        ]

        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)


        # crf层输出
        segment_output = L.Dense(segment_output_dim, **config['layer_time_distributed'])(tensor)
        crf_output = crf(segment_output)





        # label predict输出
        #predict_label_embedding = L.Dense(64)(crf_output)
        # logit_classify = embed_model.output
        # for layer in layer_stack:
        #     logit_classify = layer(logit_classify)

        logits_label = PoolerInputLayer(128,label_output_dim)([tensor,crf_output])
        #logits_label = L.Dense(label_output_dim, **config['layer_time_distributed'])(tf.concat([logit_classify,crf_output],axis=-1))
        label_predict_output = tf.nn.softmax(logits_label)


        self.tf_model = keras.Model(embed_model.inputs, [crf_output,label_predict_output])
        #self.tf_model = keras.Model(embed_model.inputs, crf_output)
        self.crf_layer = crf

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            loss = [self.crf_layer.loss,self.cascade_label_cross_entropy]
        if metrics is None:
            metrics = [[self.crf_layer.accuracy],[sparse_categorical_accuracy]]
        super(Cascade_Model, self).compile_model(loss=loss,
                                                    optimizer=optimizer,
                                                    metrics=metrics,
                                                    **kwargs)


    def predict(self,
                x_data,
                batch_size: int = 32,
                truncating: bool = False,
                predict_kwargs: Dict = None) :
        """
        Generates output predictions for the input samples.

        Computation is done in batches.

        Args:
            x_data: The input data, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
            batch_size: Integer. If unspecified, it will default to 32.
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            predict_kwargs: arguments passed to :meth:`tf.keras.Model.predict`

        Returns:
            array(s) of predictions.
        """
        if predict_kwargs is None:
            predict_kwargs = {}

        if truncating:
            seq_length = self.max_sequence_length
        else:
            seq_length = None

        if type(x_data[0]) == list:
            x_data = ["".join(x).replace("##","").replace('[CLS]',"").replace("[SEP]","") for x in x_data]
        x_data_tokenized = [self.text_processor.text_tokenizer.tokenize(x) for x in x_data]
        lengths = [len(x) for x in x_data_tokenized]

        tensor = self.text_processor.transform(x_data,seq_length=seq_length)
        (pred,pred_labels) = self.tf_model.predict(tensor, batch_size=batch_size, verbose=1, **predict_kwargs)
        #pred= self.tf_model.predict(tensor, batch_size=batch_size, verbose=1, **predict_kwargs)
        pred = pred.argmax(-1)
        pred_labels = pred_labels.argmax(-1)

        x_data_mapping = [self.text_processor.text_tokenizer.rematch(x, x_tokens) for x, x_tokens in
                          zip(x_data, x_data_tokenized)]

        res = self.label_processor.inverse_transform((pred,pred_labels),lengths=lengths,mapping_list=x_data_mapping)
        logger.debug('predict output: {}'.format(np.array(pred).shape))
        logger.debug('predict output argmax: {}'.format(pred))
        return res

    def cascade_label_cross_entropy(self, y_true, y_pred):
        from tensorflow.keras.losses import sparse_categorical_crossentropy
        from tensorflow.keras import backend as K
        import tensorflow as tf
        y_pred = tf.cast(y_pred,tf.float32)
        y_true = tf.cast(y_true,tf.float32)
        mask = K.cast(K.not_equal(y_true, 0),tf.int32)  # 将y_true 中所有为0的找出来，标记为False

        loss_ = sparse_categorical_crossentropy(y_true, y_pred)
        mask = K.cast(mask, dtype=loss_.dtype)  # 将前面统计的是否零转换成1，0的矩阵
        loss_ *= mask  # 将正常计算的loss加上mask的权重，就剔除了padding 0的影响
        loss_ = tf.math.divide_no_nan(tf.reduce_sum(loss_, axis=-1), tf.reduce_sum(mask,axis=-1))
        #loss_ = tf.reduce_mean(sparse_categorical_crossentropy(y_true, y_pred))
        return loss_#K.mean(loss_)



from bert4keras.models import build_transformer_model
from bert4keras.backend import keras, K
from bert4keras.optimizers import Adam
from bert4keras.optimizers import extend_with_weight_decay
from bert4keras.optimizers import extend_with_layer_adaptation
from bert4keras.optimizers import extend_with_piecewise_linear_lr
from bert4keras.optimizers import extend_with_gradient_accumulation
from keras.layers import Input, Lambda
from keras.models import Model

from tensorflow.keras import  backend as K


floatx = K.floatx()

def build_transformer_model_with_mlm(config_path):
    """带mlm的bert模型
    """
    bert = build_transformer_model(
        config_path, with_mlm='linear', return_keras_model=False
    )
    proba = bert.model.output

    # 辅助输入
    token_ids = Input(shape=(None,), dtype='int64', name='token_ids')  # 目标id
    is_masked = Input(shape=(None,), dtype=floatx, name='is_masked')  # mask标记

    def mlm_loss(inputs):
        """计算loss的函数，需要封装为一个层
        """
        y_true, y_pred, mask = inputs
        loss = K.sparse_categorical_crossentropy(
            y_true, y_pred, from_logits=True
        )
        loss = K.sum(loss * mask) / (K.sum(mask) + K.epsilon())
        return loss

    def mlm_acc(inputs):
        """计算准确率的函数，需要封装为一个层
        """
        y_true, y_pred, mask = inputs
        y_true = K.cast(y_true, floatx)
        acc = keras.metrics.sparse_categorical_accuracy(y_true, y_pred)
        acc = K.sum(acc * mask) / (K.sum(mask) + K.epsilon())
        return acc

    mlm_loss = Lambda(mlm_loss, name='mlm_loss')([token_ids, proba, is_masked])
    mlm_acc = Lambda(mlm_acc, name='mlm_acc')([token_ids, proba, is_masked])

    train_model = Model(
        bert.model.inputs + [token_ids, is_masked], [mlm_loss, mlm_acc]
    )

    loss = {
        'mlm_loss': lambda y_true, y_pred: y_pred,
        'mlm_acc': lambda y_true, y_pred: K.stop_gradient(y_pred),
    }

    return bert, train_model, loss


def build_transformer_model_with_lm(config_path):
    """带lm的bert模型
    """
    bert = build_transformer_model(
        config_path,
        with_mlm='linear',
        application='lm',
        return_keras_model=False
    )
    token_ids = bert.model.inputs[0]
    proba = bert.model.output

    def lm_loss(inputs, mask=None):
        """计算loss的函数，需要封装为一个层
        """
        y_true, y_pred = inputs
        y_true, y_pred = y_true[:, 1:], y_pred[:, :-1]

        if mask is None:
            mask = 1.0
        else:
            mask = K.cast(mask[1][:, 1:], floatx)

        loss = K.sparse_categorical_crossentropy(
            y_true, y_pred, from_logits=True
        )
        loss = K.sum(loss * mask) / (K.sum(mask) + K.epsilon())
        return loss

    def lm_acc(inputs, mask=None):
        """计算准确率的函数，需要封装为一个层
        """
        y_true, y_pred = inputs
        y_true, y_pred = K.cast(y_true[:, 1:], floatx), y_pred[:, :-1]

        if mask is None:
            mask = 1.0
        else:
            mask = K.cast(mask[1][:, 1:], floatx)

        acc = keras.metrics.sparse_categorical_accuracy(y_true, y_pred)
        acc = K.sum(acc * mask) / (K.sum(mask) + K.epsilon())
        return acc

    lm_loss = Lambda(lm_loss, name='lm_loss')([token_ids, proba])
    lm_acc = Lambda(lm_acc, name='lm_acc')([token_ids, proba])

    train_model = Model(bert.model.inputs, [lm_loss, lm_acc])

    loss = {
        'lm_loss': lambda y_true, y_pred: y_pred,
        'lm_acc': lambda y_true, y_pred: K.stop_gradient(y_pred),
    }

    return bert, train_model, loss


def build_transformer_model_with_unilm(config_path):
    """带unilm的bert模型
    """
    bert = build_transformer_model(
        config_path,
        with_mlm='linear',
        application='unilm',
        return_keras_model=False
    )
    token_ids = bert.model.inputs[0]
    segment_ids = bert.model.inputs[1]
    proba = bert.model.output

    def unilm_loss(inputs, mask=None):
        """计算loss的函数，需要封装为一个层
        """
        y_true, y_pred, segment_ids = inputs
        y_true, y_pred = y_true[:, 1:], y_pred[:, :-1]

        if mask is None:
            mask = 1.0
        else:
            mask = K.cast(mask[1][:, 1:], floatx)

        segment_ids = K.cast(segment_ids, floatx)
        mask = mask * segment_ids[:, 1:]

        loss = K.sparse_categorical_crossentropy(
            y_true, y_pred, from_logits=True
        )
        loss = K.sum(loss * mask) / (K.sum(mask) + K.epsilon())
        return loss

    def unilm_acc(inputs, mask=None):
        """计算准确率的函数，需要封装为一个层
        """
        y_true, y_pred, segment_ids = inputs
        y_true, y_pred = K.cast(y_true[:, 1:], floatx), y_pred[:, :-1]

        if mask is None:
            mask = 1.0
        else:
            mask = K.cast(mask[1][:, 1:], floatx)

        segment_ids = K.cast(segment_ids, floatx)
        mask = mask * segment_ids[:, 1:]

        acc = keras.metrics.sparse_categorical_accuracy(y_true, y_pred)
        acc = K.sum(acc * mask) / (K.sum(mask) + K.epsilon())
        return acc

    token_proba_segment = [token_ids, proba, segment_ids]
    unilm_loss = Lambda(unilm_loss, name='unilm_loss')(token_proba_segment)
    unilm_acc = Lambda(unilm_acc, name='unilm_acc')(token_proba_segment)

    train_model = Model(bert.model.inputs, [unilm_loss, unilm_acc])

    loss = {
        'unilm_loss': lambda y_true, y_pred: y_pred,
        'unilm_acc': lambda y_true, y_pred: K.stop_gradient(y_pred),
    }

    return bert, train_model, loss


def build_transformer_model_for_pretraining(model,model_config,model_ckpt):
    """构建训练模型，通用于TPU/GPU
    注意全程要用keras标准的层写法，一些比较灵活的“移花接木”式的
    写法可能会在TPU上训练失败。此外，要注意的是TPU并非支持所有
    tensorflow算子，尤其不支持动态（变长）算子，因此编写相应运算
    时要格外留意。
    """
    learning_rate = 0.00176
    weight_decay_rate = 0.01
    grad_accum_steps = 16  # 大于1即表明使用梯度累积
    which_optimizer = 'lamb'  # adam 或 lamb，均自带weight decay
    exclude_from_weight_decay = ['Norm', 'bias']
    exclude_from_layer_adaptation = ['Norm', 'bias']
    num_train_steps = 125000
    num_warmup_steps = 3125

    lr_schedule = {
        num_warmup_steps * grad_accum_steps: 1.0,
        num_train_steps * grad_accum_steps: 0.0,
    }

    if model == 'roberta':
        bert, train_model, loss = build_transformer_model_with_mlm(model_config)
    elif model == 'gpt':
        bert, train_model, loss = build_transformer_model_with_lm(model_config)
    elif model == 'unilm':
        bert, train_model, loss = build_transformer_model_with_unilm(model_config)
    else:
        raise ValueError("Not support")

    # 优化器
    optimizer = extend_with_weight_decay(Adam)
    if which_optimizer == 'lamb':
        optimizer = extend_with_layer_adaptation(optimizer)
    optimizer = extend_with_piecewise_linear_lr(optimizer)
    optimizer_params = {
        'learning_rate': learning_rate,
        'lr_schedule': lr_schedule,
        'weight_decay_rate': weight_decay_rate,
        'exclude_from_weight_decay': exclude_from_weight_decay,
        'exclude_from_layer_adaptation': exclude_from_layer_adaptation,
        'bias_correction': False,
    }
    if grad_accum_steps > 1:
        optimizer = extend_with_gradient_accumulation(optimizer)
        optimizer_params['grad_accum_steps'] = grad_accum_steps
    optimizer = optimizer(**optimizer_params)

    # 模型定型
    train_model.compile(loss=loss, optimizer=optimizer)

    # 如果传入权重，则加载。注：须在此处加载，才保证不报错。
    if model_ckpt is not None:
        bert.load_weights_from_checkpoint(model_ckpt)

    return bert,train_model


from typing import Tuple, List
from nlp_tools import macros as k
import os
CORPUS_PATH = os.path.join(k.DATA_PATH, 'corpus')

class DataReader(object):

    @staticmethod
    def read_conll_format_file(file_path: str,
                               text_index: int = 0,
                               label_index: int = 1) -> Tuple[List[List[str]], List[List[str]]]:
        """
        Read conll format data_file
        Args:
            file_path: path of target file
            text_index: index of text data, default 0
            label_index: index of label data, default 1

        Returns:

        """
        x_data, y_data = [], []
        with open(file_path, 'r', encoding='utf-8') as f:
            lines = f.read().splitlines()
            x, y = [], []
            for line in lines:
                rows = line.split(' ')
                if len(rows) == 1:
                    x_data.append(x)
                    y_data.append(y)
                    x = []
                    y = []
                else:
                    x.append(rows[text_index])
                    y.append(rows[label_index])
        return x_data, y_data

from nlp_tools.corpus.ner.corpus_loader import ChineseDailyNerCorpus
from nlp_tools.corpus.classify.corpus_loader import SMP2018ECDTCorpus
from nlp_tools.corpus.classify.dialogue_corpus_loader import DialogueCorpusLoader


from abc import ABC
import numpy as np
from typing import List,Dict,Any,Union

from sklearn import metrics as sklearn_metrics
from tensorflow import keras

import nlp_tools
from nlp_tools.embeddings import ABCEmbedding,BareEmbedding
from nlp_tools.generators import CorpusGenerator
from nlp_tools.layers import L
from nlp_tools.logger import logger
from nlp_tools.processors import ClassificationProcessor
from nlp_tools.processors import SequenceProcessor
from nlp_tools.tasks.abs_task_model import  ABCTaskModel
from nlp_tools.types import TextSamplesVar,ClassificationLabelVar
from nlp_tools.generators import CorpusGenerator, BatchGenerator

from nlp_tools.utils.training_tips import creat_FGM
import functools

class ABCSimilaryModel(ABCTaskModel,ABC):
    '''
        Abstract Similary Model
    '''
    __task__ = 'similary'

    def to_dict(self) -> Dict:
        info = super(ABCSimilaryModel,self).to_dict()
        info['config']['use_FGM'] = self.use_FGM
        info['config']['subtask'] = self.subtask
        return info

    def __init__(self,
                 embedding,
                 text_processor,
                 label_processor,
                 sequence_length: int = None,
                 hyper_parameters:Dict[str,Dict[str,Any]] = None,
                 subtask='Classification',
                 use_FGM=True,
                 train_sequece_length_as_max_sequence_length=None,
                 max_sequence_length=None):
        """
                        Args:
                            embedding: embedding object
                            sequence_length: target sequence length
                            hyper_parameters: hyper_parameters to overwrite
                            multi_label: is multi-label classification
                            text_processor: text processor
                            label_processor: label processor
                        """
        super(ABCSimilaryModel,self).__init__()


        if hyper_parameters is None:
            hyper_parameters = self.default_hyper_parameters()

        self.tf_model:keras.Model = None
        self.embedding = embedding
        self.hyper_parameters = hyper_parameters
        self.sequence_length = sequence_length

        self.text_processor = text_processor
        self.label_processor = label_processor

        self.subtask = subtask
        self.text_processor = text_processor
        self.label_processor = label_processor

        self.max_sequence_length = max_sequence_length
        self.train_max_sequence_length = None
        self.train_sequece_length_as_max_sequence_length = train_sequece_length_as_max_sequence_length
        self.use_FGM = use_FGM

        ## 如果分词器支持keep_tokens，且embedding也支持keep_tokens，则需要吧分词器的keep_tokens跟新到
        if self.text_processor:
            if hasattr(self.text_processor.text_tokenizer,
                       "keep_tokens") and self.text_processor.text_tokenizer.keep_tokens != "" and hasattr(
                    self.embedding, "keep_tokens"):
                self.embedding.keep_tokens = self.text_processor.text_tokenizer.keep_tokens

        self.embedding_max_position = None
        if self.embedding and hasattr(self.embedding, "max_position"):
            self.embedding_max_position = self.embedding.max_position

    def build_model(self,
                    x_train:TextSamplesVar,
                    y_train:ClassificationLabelVar) -> None:
        """
                       Build Model with x_data and y_data
                       This function will setup a :class:`CorpusGenerator`,
                        then call py:meth:`ABCClassificationModel.build_model_gen` for preparing processor and model
                       Args:
                           x_train:
                           y_train:
                       Returns:
                       """
        train_gen = CorpusGenerator(x_train,y_train)
        self.build_model_generator([train_gen])

    def build_model_generator(self,
                              generators: List[CorpusGenerator]) -> None:

        self.label_processor.build_vocab_generator(generators)
        if self.train_max_sequence_length is None:
            self.train_max_sequence_length = self.text_processor.get_max_seq_length_from_corpus(generators,
                                                                                                cover_rate=1.0)

        if self.train_sequece_length_as_max_sequence_length:
            self.max_sequence_length = self.train_sequece_length_as_max_sequence_length

        #self.label_processor.update_length_info(embedding_max_position=self.embedding_max_position,
        #                                       max_sentence_length=self.max_sequence_length)
        self.text_processor.update_length_info(embedding_max_position=self.embedding_max_position,
                                               max_sentence_length=self.max_sequence_length)

        self.embedding.build_embedding_model()

        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

            # 替换model.train_step 方法即可,并且删除原有的 train_function方法

            if self.use_FGM:
                # if self.tf_model.train_function is not None:
                #    self.tf_model.make_train_function()
                # old_train_function = self.tf_model.train_function
                train_step = creat_FGM()
                self.tf_model.train_step = functools.partial(train_step, self.tf_model)
                self.tf_model.train_function = None

    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss:Any = None,
                      optimizer: Any=None,
                      metrics:Any=None,
                      **kwargs:Any) -> None:
        """
                        Configures the model for training.
                        call :meth:`tf.keras.Model.predict` to compile model with custom loss, optimizer and metrics
                        Examples:
                            >>> model = BiLSTM_Model()
                            # Build model with corpus
                            >>> model.build_model(train_x, train_y)
                            # Compile model with custom loss, optimizer and metrics
                            >>> model.compile(loss='categorical_crossentropy', optimizer='rsm', metrics = ['accuracy'])
                        Args:
                            loss: name of objective function, objective function or ``tf.keras.losses.Loss`` instance.
                            optimizer: name of optimizer or optimizer instance.
                            metrics (object): List of metrics to be evaluated by the model during training and testing.
                            **kwargs: additional params passed to :meth:`tf.keras.Model.predict``.
                        """
        if loss is None:
            loss = 'sparse_categorical_crossentropy'

        if optimizer is None:
            optimizer = 'adam'

        if metrics is None:
            metrics=['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)


    def fit(self,
            train_data,
            validate_data=None,
            batch_size: int=64,
            epochs:int = 5,
            callbacks:List['keras.callbacks.Callback'] =None,
            fit_kwargs:Dict=None,
            generator=None) ->'keras.callbacks.History':
        if generator == None:
            generator = BatchGenerator

        train_generator = generator(train_data,
                                    text_processor=self.text_processor,
                                    label_processor=self.label_processor,
                                    seq_length=self.train_max_sequence_length,
                                    batch_size=batch_size)
        if validate_data:
            valid_generator = generator(validate_data,
                                        text_processor=self.text_processor,
                                        label_processor=self.label_processor,
                                        seq_length=self.train_max_sequence_length,
                                        batch_size=batch_size)
        else:
            valid_generator = None

        return self.fit_generator(train_sample_gen= train_generator,
                                  valid_sample_gen = valid_generator,
                                  batch_size=batch_size,
                                  epochs = epochs,
                                  callbacks=callbacks,
                                  fit_kwargs=fit_kwargs)

    def fit_generator(self,
                      train_sample_gen,
                      valid_sample_gen,
                      *,
                      batch_size:int=64,
                      epochs:int=5,
                      callbacks:List['keras.callbacks.Callback'] = None,
                      fit_kwargs:Dict=None) -> 'keras.callbacks.History':
        """
                        Trains the model for a given number of epochs with given data generator.
                        Data generator must be the subclass of `CorpusGenerator`
                        Args:
                            train_sample_gen: train data generator.
                            valid_sample_gen: valid data generator.
                            batch_size: Number of samples per gradient update, default to 64.
                            epochs: Number of epochs to train the model.
                                An epoch is an iteration over the entire `x` and `y` data provided.
                            callbacks: List of `tf.keras.callbacks.Callback` instances.
                                List of callbacks to apply during training.
                                See `tf.keras.callbacks`.
                            fit_kwargs: fit_kwargs: additional arguments passed to :meth:`tf.keras.Model.fit`
                        Returns:
                            A :py:class:`tf.keras.callback.History`  object. Its `History.history` attribute is
                            a record of training loss values and metrics values
                            at successive epochs, as well as validation loss values
                            and validation metrics values (if applicable).
                        """
        if fit_kwargs == None:
            fit_kwargs = {}

        self.build_model_generator([g for g in [train_sample_gen, valid_sample_gen] if g])
        return self.tf_model.fit(train_sample_gen.forfit(),
                                 steps_per_epoch=len(train_sample_gen),
                                 validation_data=valid_sample_gen.forfit(random=False),
                                 validation_steps=len(valid_sample_gen),
                                 epochs=epochs,
                                 callbacks=callbacks,
                                 **fit_kwargs)

    def predict(self,
                x_data:TextSamplesVar,
                *,
                batch_size:int=32,
                truncating:bool=False,
                label_threshold:float = 0.5,
                predict_kwargs:Dict = None) -> ClassificationLabelVar:
        """
                        Generates output predictions for the input samples.
                        Computation is done in batches.
                        Args:
                            x_data: The input data, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
                            batch_size: Integer. If unspecified, it will default to 32.
                            truncating: remove values from sequences larger than `model.embedding.sequence_length`
                            multi_label_threshold:
                            predict_kwargs: arguments passed to ``predict()`` function of ``tf.keras.Model``
                        Returns:
                            array(s) of predictions.
                        """
        if predict_kwargs is None:
            predict_kwargs = {}

        with nlp_tools.utils.custom_object_scope():
            if truncating:
                seq_length = self.sequence_length
            else:
                seq_length = None
            tensor = self.text_processor.transform(x_data,
                                                   segment=self.embedding.segment,
                                                   seq_length=seq_length,
                                                   max_position=self.embedding.max_position)
            logger.debug(f'predict input shape {np.array(tensor).shape} x: \n{tensor}')
            pred = self.tf_model.predict(tensor, batch_size=batch_size, **predict_kwargs)
            logger.debug(f'predict output shape {pred.shape}')
            pred_argmax = pred.argmax(-1)
            lengths = [len(sen) for sen in x_data]
            res = self.label_processor.inverse_transform(pred_argmax, lengths=lengths)
            logger.debug(f'predict output argmax: {pred_argmax}')
        return res

    def evaluate(self,
                 x_data: TextSamplesVar,
                 y_data: ClassificationLabelVar,
                 *,
                 batch_size: int = 32,
                 digits: int = 4,
                 label_threshold: float = 0.5,
                 truncating: bool = False) -> Dict:
        y_pred = self.predict(x_data,
                              batch_size=batch_size,
                              truncating=truncating,
                              label_threshold=label_threshold)


        original_report = sklearn_metrics.classification_report(y_data,
                                                                y_pred,
                                                                output_dict=True,
                                                                digits=digits)

        print(sklearn_metrics.classification_report(y_data,
                                                    y_pred,
                                                    output_dict=False,
                                                    digits=digits))
        report = {
            'detail': original_report,
            **original_report['weighted avg']
        }
        return report


    def make_output_layer(self,num_classes=0) -> keras.layers.Layer:
        """:return: a correctly shaped keras dense layer for model output."""
        if self.subtask == "Classification":
            return keras.layers.Dense(num_classes, activation='softmax')
        elif self.subtask == "Ranking":
            return keras.layers.Dense(1, activation='linear')
        else:
            raise ValueError(f"{self.subtask} is not a valid task type."
                             f"Must be in `Ranking` and `Classification`.")




